/*
  Last changed Time-stamp: <2010-11-10 23:59:40 xtof>
  $Id: Alignment.cpp,v 1.5 2010/11/10 23:21:01 xtof Exp $
*/
#include <algorithm>
#include <functional>
#include <iostream>
#include <iterator>
#include <map>
#include <string>
#include <utility>
#include <vector>

#include "Maps.h"
#include "TypeDefs.h"
#include "Utils.h"

extern RandomNumber rnd;

// private function(s)
static int
count_cyclic_splits(std::string column);
static char
column_to_type(std::string column);
static std::string
filter_out_missing(std::string source, CharIntMap_t missing);
static std::string
filter_out_gaps(std::string source);

/**
 * Converts vector of alignment rows to vector of alignment columns
 */
std::vector<std::string>
listOf_rows_to_listOf_columns (std::vector<std::string> rows,
			       std::vector<int> order)
{
  std::vector<std::string> columns = std::vector<std::string>();

  for (unsigned int column = 0; column < rows[0].size(); column++) {
    std::string str = std::string();
    for (unsigned int row = 0; row < rows.size(); row++) {
      str.push_back(rows[order[row]][column]);
    }
    columns.push_back(str);
  }

  return (columns);
}

/**
 * Converts a vector of alignment columns to a vector of alignment rows
 */
std::vector<std::string>
listOf_columns_to_listOf_rows (std::vector<std::string> columns)
{
  std::vector<std::string> rows = std::vector<std::string>(columns[0].size(),
							   std::string());

  for (unsigned int column = 0; column < columns.size(); column++) {
    for (unsigned int row = 0; row < columns[0].size(); row++) {
      rows[row].push_back(columns[column][row]);
    }
  }

  return (rows);
}

/**
 *
 */
Statistic_t
shuffle_each_column(std::vector<std::string> columns,
		    int times, CharIntMap_t missing,
		    int nogaps, int scilent, char *inname)
{
  std::string classify_me = std::string();     
  std::string c_types = std::string();
  std::vector<double> p_values = std::vector<double>();
  int cnt = 0; // for progress bar
  
#ifdef _DEBUG_
  int k = 0;
#endif
  
  // foreach column ...
  for (std::vector<std::string>::const_iterator column = columns.begin();
      column != columns.end();
      column++) {
    double counter = 0;

    // progress bar
    if ( ! scilent ) {
      char chars[5] = "|/-\\";
      pretty_print(stderr, "\b\b\b\b\b\b");
      pretty_print(stderr, "%.2f %c",
		   static_cast<float>(cnt)/static_cast<float>(columns.size()),
		   chars[cnt % 4]);
      cnt++;
      fflush(stderr);
    }

    // remove missing data
    std::string without_missing = filter_out_missing(*column, missing);

    if (nogaps == 1) {
      classify_me = filter_out_gaps(without_missing);
    }
    else {
      classify_me = without_missing;
    }

    // classify column
    char c_type = column_to_type(classify_me);
    c_types.push_back(c_type);
    
    // all rows are missing characters
    if (without_missing.empty()) {
      p_values.push_back(0.0);

#ifdef _DEBUG_
      pretty_print(stdout, "%-15s #%d %c orig column SKIPPED\n",
		   (*column).c_str(), k++, c_type);
#endif
 
      continue;
    }

    // one row is a character or gap
    if (without_missing.size() == 1) {
      p_values.push_back(0.0);

#ifdef _DEBUG_
      pretty_print(stdout, "%-15s #%d %c orig column SKIPPED\n",
		   (*column).c_str(), k++, c_type);
#endif

      continue;
    }

    if ((c_type == 'C') ||(c_type == 'S')) {
      p_values.push_back(0.0);

#ifdef _DEBUG_
      pretty_print(stdout, "%-15s #%d %c orig column SKIPPED\n",
		   (*column).c_str(), k++, c_type);
#endif

      continue;
    }

#ifdef _DEBUG_
    pretty_print(stdout, "%-15s #%d %c orig column\n",
		 (*column).c_str(), k++, c_type);
#endif

    // count number of splits in real data
    int splits_real = count_cyclic_splits(without_missing);
  
#ifdef _DEBUG_
    pretty_print(stdout, "%-15s %d reduced\n",
		 without_missing.c_str(), splits_real);
#endif  

    // random shuffle column many times ...
    for (int i = 0; i<times; i++) {

      std::string tmp(without_missing);
      random_shuffle(tmp.begin(), tmp.end(), rnd);
      int splits_shuffled = count_cyclic_splits(tmp);

      // ... and count significant events ...     
      if (splits_real < splits_shuffled) counter++;
    }
    p_values.push_back(counter/times);

  }

  if ( ! scilent )
    std::cerr << "\b\b\b\b\b\b" << inname << " processed" << std::endl;
  
  return(make_pair(p_values,c_types));
}

std::vector<double>
smooth_values(std::vector<double> p_values, int smooth)
{
  // nothing to smooth here!
  if (smooth <= 1) return (std::vector<double>());
  
  std::vector<double> smoothed = std::vector<double>(p_values.size(), 0.0);
  
  for (unsigned int idx=0; idx<(p_values.size()-smooth-1); idx++) {
    double mean = 0.0;
    for (unsigned int i=idx; i<(idx+smooth); i++)
      mean += p_values[i];

    mean /= smooth;

    for (unsigned int i=idx; i<(idx+smooth); i++)
      smoothed[i] = (p_values[i] > mean) ? p_values[i] : mean;

  }

  return (smoothed);
}

/**
 *
 */
std::vector<std::string>
to_cyclic_order(std::vector<std::string> orig, std::vector<int> order)
{
  std::vector<std::string> reordered = std::vector<std::string>();
  for (unsigned int i=0; i<orig.size(); i++)
    reordered.push_back(orig[order[i]]);

  return (reordered);
}

typedef std::pair<char,double> Duple_t;
typedef std::pair<std::string, Duple_t> Tuple_t;

// function object
struct is_random
  : std::binary_function<ColumnData_t, double, bool> {
  bool operator()(const ColumnData_t item, const double cutoff) const {
    char c_type    = item.c_type;
    double p_value = item.p_value;
    if ((p_value <= 0.4 && p_value >= 0.6) || (c_type == 'C')) return (false);
    else return (true);
  }
};

// function object
struct is_lt_cutoff_keep_const_columns
  : std::binary_function<ColumnData_t, double, bool> {
  bool operator()(const ColumnData_t item, const double cutoff) const {
    char c_type    = item.c_type;
    double p_value = item.p_value;
    if ((p_value >= cutoff) || (c_type == 'C')) return (false);
    else return (true);
  }
};

// function object
struct is_lt_cutoff
  : std::binary_function<ColumnData_t, double, bool> {
  bool operator()(const ColumnData_t item, const double cutoff) const {
    double p_value = item.p_value;
    if (p_value >= cutoff) return (false);
    else return (true);
  }
};

// function object
struct is_gt_cutoff
  : std::binary_function<Tuple_t, double, bool> {
  bool operator()(const Tuple_t item, const double cutoff) const {
    double p_value = item.second.second;
    char c_type    = item.second.first;
    if (c_type == 'C') {//std::cerr << "removing C\n";
      return (true);}
    if (c_type == 'N') {//std::cerr << "removing N\n";
      return (true);}
    if (c_type == 'S') {//std::cerr << "removing S\n";
      return (true);}
    if (p_value > cutoff) {//std::cerr << "removing >" << p_value << "<\n";
      return (true);}
    else {//std::cerr << "keeping " << c_type << " >" << p_value << "<\n";
      return (false);}
  }
};

// function object
struct is_not_variable
  : std::binary_function<Tuple_t, double, bool> {
  bool operator()(const Tuple_t item, const double cutoff) const {
    // double p_value = item.second.second;
    char c_type    = item.second.first;
    if (c_type == 'C') {// std::cerr << "removing C\n";
      return (true);}
    if (c_type == 'N') {//std::cerr << "removing N\n";
      return (true);}
    if (c_type == 'S') {//std::cerr << "removing S\n";
      return (true);}
    else {//std::cerr << "keeping " << c_type << " >" << p_value << "<\n";
      return (false);}
  }
};

// function object
struct extract_column
  : std::unary_function<ColumnData_t, std::string> {
  std::string operator()(const ColumnData_t item) const {
    return (item.column);
  }
};

// function object
struct extract_column_c_type_pair
  : std::unary_function<Tuple_t, std::pair<std::string,char> > {
  std::pair<std::string,char> operator()(const Tuple_t item) const {
    return (make_pair(item.first, item.second.first));
  }
};

// function object
struct extract_column_p_value_pair
  : std::unary_function<Tuple_t, std::pair<std::string,double> > {
  std::pair<std::string,double> operator()(const Tuple_t item) const {
    return (make_pair(item.first, item.second.second));
  }
};

std::vector<Tuple_t>
make_array_of_pair_of_pairs(std::vector<std::string> col,
			    Statistic_t stat) {
  std::vector<std::string> cols = col;
  std::vector<double> p_values = stat.first;
  std::string c_types = stat.second;
  std::vector<Tuple_t > ts = std::vector<Tuple_t >();

  // How about transform in transform? seems to be more elegant
  for(unsigned int i=0; i<cols.size(); i++) {
    Duple_t d = std::pair<char,double>(c_types[i], p_values[i]);
    Tuple_t t = std::pair<std::string,Duple_t>(cols[i], d);
    ts.push_back(t);
  }

  return(ts);
}

double
relative_frequency_of_random_columns(std::vector<std::string> col,
				     Statistic_t stat,
				     double cutoff)
{
  std::vector<std::string> cols = col;
  std::vector<double> p_values = stat.first;
  std::string c_types = stat.second;
  std::vector<Tuple_t > ts = make_array_of_pair_of_pairs(col, stat);
  std::vector<Tuple_t > ts1 = make_array_of_pair_of_pairs(col, stat);

  ts.erase(std::remove_if(ts.begin(), ts.end(),
			  std::bind2nd(is_gt_cutoff(), 0.5)),
	   ts.end());

  // std::cerr << "is_gt_cutoff(): " << ts.size() << std::endl;
 
  ts1.erase(std::remove_if(ts1.begin(), ts1.end(),
			  std::bind2nd(is_not_variable(), 0.5)),
	   ts1.end());

  // std::cerr << "is_not_variable(): " << ts1.size() << std::endl;

  // relative frequency of random positions in alignment
  double rf = static_cast<double>(ts.size());
  rf /= static_cast<double>(ts1.size());
  
  return (rf);
}

Result_t
filter_columns(std::vector<std::string> col, Statistic_t stat,
	       double cutoff, int noconst)
{
  // alocate memory
  std::vector<ColumnData_t> colData(col.size());
  std::vector<std::string> remains(col.size());
  std::vector<int> keep_idx(colData.size(),-1);
  
  // construct column data vector
  for(unsigned int i=0; i<col.size(); i++) {
    colData[i].column = col[i];
    colData[i].c_type = stat.second[i];
    colData[i].c_idx = i;
    colData[i].p_value = stat.first[i];
  }

  //// DEBUG
#ifdef _DEBUG_
  std::cout << "!!! colData BEFORE filtering !!!" << std::endl;
  std::cout << "colData.size() = " << colData.size()<< std::endl;
  for (std::vector<ColumnData_t>::iterator it = colData.begin();
       it != colData.end();
       it++) {
    std::cout << it->c_idx << "|"
              << it->column << "|"
	      << it->c_type << "|"
	      << it->p_value  << "|"
	      << std::endl;
  }
  std::cout << "~~~~~~~~~~" << std::endl;
#endif
  ////

  // filter out columns according to userspecified settings
  if (noconst == 1)
    colData.erase(std::remove_if(colData.begin(), colData.end(),
			   std::bind2nd(is_lt_cutoff(), cutoff)),
	     colData.end());
  else if (noconst == 2)
    colData.erase(std::remove_if(colData.begin(), colData.end(),
			    std::bind2nd(is_random(), cutoff)),
	     colData.end());   
  else
    colData.erase(std::remove_if(colData.begin(), colData.end(),
			    std::bind2nd(is_lt_cutoff_keep_const_columns(),
					 cutoff)),
	     colData.end());

  //// DEBUG
#ifdef _DEBUG_
  std::cout << "!!! colData AFTER filtering !!!" << std::endl;
  std::cout << "colData.size() = " << colData.size()<< std::endl;
  for (std::vector<ColumnData_t>::iterator it = colData.begin();
       it != colData.end();
       it++) {
    std::cout << it->c_idx << "|"
              << it->column << "|"
	      << it->c_type << "|"
	      << it->p_value  << "|"
	      << std::endl;
  }
  std::cout << "~~~~~~~~~~" << std::endl;
#endif
  ////

  // shrink memory of result to appropriate size
 remains.resize(colData.size());

 // extract columns from remains
 std::transform(colData.begin(),colData.end(),remains.begin(),extract_column());

 // extract kept indices
 for (std::vector<ColumnData_t>::iterator it = colData.begin();
      it != colData.end();
      it++) {
   keep_idx[it->c_idx] = it->c_idx;
 }

 //// DEBUG
#ifdef _DEBUG_
 std::cout << "remains.size() = " << remains.size() << std::endl;
 for (std::vector<std::string>::iterator it = remains.begin();
      it != remains.end();
      it++) {
   std::cout << *it << std::endl;
 }
 std::cout << "remains.size() = " << remains.size() << std::endl;
 std::cout << "~~~~~~~~~~" << std::endl;
#endif
 ////
 
 return (make_pair(remains,keep_idx));
}

/**
 *
 */
static int
count_cyclic_splits(std::string column)
{
  std::string col = column.substr(column.size()-1,1);
  col += column;

  int counter = 0;
  for (unsigned int i = 0; i<col.size()-1; i++)
    if (col[i] != col[i+1]) counter++;

#ifdef _DEBUG_
  pretty_print(stdout, "%-15s %d cyclic count\n", col.c_str(), counter);
#endif 

  return (counter);
}

/**
 *
 */
static char
column_to_type(std::string column)
{
  if (column.empty()) return('N');
  
  // uniq-count column symbols
  CharIntMap_t m
    = CharIntMap_new_from_char_array(const_cast<char*>(column.c_str()));
  std::string keys = CharIntMap_keys(m);

#ifdef _DEBUG_
  std::cout << "|" ;
  std::copy(keys.begin(),keys.end(),
	    std::ostream_iterator<char>(std::cout, " "));
  std::cout << "|" << std::endl;
#endif
  
  // constant column
  if (keys.size() == 1) return ('C');

  // distinguish between parsimonious informative column and
  // singleton column with exactly two different symbols
  if (keys.size() == 2) {
    if ((m[keys[0]] > 1) && (m[keys[1]] > 1)) return ('P');
    return ('S');
  }

  return ('M');
}

// function object
struct is_missing_data
  : std::binary_function<char, std::map<char,int>, bool> {
  bool operator()(const char key, const CharIntMap_t m) const {
    return (CharIntMap_exists(key, m));
  }
};


/**
 *
 */
static std::string
filter_out_missing(std::string source, CharIntMap_t missing)
{
  std::string str = source;
  std::string::iterator it
    = remove_if(str.begin(), str.end(),
		std::bind2nd(is_missing_data(), missing));
  str.erase(it, str.end());

  return (str);
}

/**
 *
 */
static std::string
filter_out_gaps(std::string source)
{
  
#ifdef _DEBUG_
  pretty_print(stdout, ">%s<<<\n", source.c_str());
#endif

  if (source.empty()) return (std::string());
  std::string str = source;
  std::string::iterator it
    = remove_if(str.begin(), str.end(),
		std::bind2nd(std::equal_to<char>(),'-'));
  str.erase(it, str.end());

#ifdef _DEBUG_
  pretty_print(stdout, ">%s<\n", str.c_str());
#endif
  
  return (str);
}

// End of file
