/*
  Last changed Time-stamp: <2011-04-19 18:22:18 xtof>
  $Id: CyclicOrderer.cpp,v 1.4 2011/04/19 16:34:37 xtof Exp $
*/
#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include <iterator>
#include <algorithm>
#include <cmath>

#include "FileFormats.h"
#include "TypeDefs.h"
#include "holders.h"
#include "Utils.h"

#define DEBUG 0

#define LINNORMALIZE 1
#define LOGNORMALZE 0

using namespace std;

static QUARTET
calcQuartetWeights (vector<string> sequences, std::string basename,
		    int weights, int silent);

void
print_quartet_to_fstream(std::ofstream& file,
			 int a, int b, int c, int d,
			 double w1, double w2, double w3)
{
  file << "quartet: "
       << format("%03d", a) << " "
       << format("%03d", b) << " "
       << format("%03d", c) << " "
       << format("%03d", d) << " "
       << " weights: "
       << w1 << " "
       << w2 << " "
       << w3 << std::endl;
}

int
get_size_ordered_idx(QUARTET & q, int a, int b, int c, int d)
{
  int x = a;
  int y = b;
  int u = c;
  int v = d;
  int m;

  if (y > x) {
    m = x;
    x = y;
    y = m;
  }
  
  if (u > x) {
    m = x;
    x = u;
    u = m;
  }
  
  if (v > x) {
    m = x;
    x = v;
    v = m;
  }
  
  if (u > y) {
    m = y;
    y = u;
    u = m;
  }
  
  if (v > y) {
    m = y;
    y = v;
    v = m;
  }
  
  if (v > u) {
    m = u;
    u = v;
    v = m;
  }

  /**
     Next, use size-ordered numbers to access the triple
  */
  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);

  if (idx >= static_cast<int>(q.size())) {
    cout << "Out of range error!" << endl;
    exit (EXIT_FAILURE);
  }

  return (idx);
}

void print_QuartetWeights_to_file(QUARTET & q, string basename, int n)
{
  ofstream file;
  file.open((basename+"_wgts.txt").c_str());
  
  for (int a=1; a<n-2; a++) {
    for (int b=a+1; b<n-1; b++) {
      for (int c=b+1; c<n; c++) {
	for (int d=c+1; d<n+1; d++) {
	  int idx = get_size_ordered_idx(q, a, b, c, d);
	  DOUBLETRIPLET w = q[idx];
	  file << "quartet: "
	       << format("%03d", a) << " "
	       << format("%03d", b) << " "
	       << format("%03d", c) << " "
	       << format("%03d", d) << " "
	       << " weights: "
	       << w.A << " "
	       << w.B << " "
	       << w.C << endl;
	}
      }
    }
  }

  file.close();
}

void normalize (QUARTET & quartet)
{
  for (unsigned int n=0; n<quartet.size (); n++) 
    {
      DOUBLETRIPLET t=quartet[n];
    
      double m=min (t.A, min (t.B, t.C));
    
      t.A-=m;
      t.B-=m;
      t.C-=m;
    
      quartet[n]=t;
    }
}

void logNormalize (QUARTET & quartet)
{
  for (unsigned int n=0; n<quartet.size (); n++) 
    {
      DOUBLETRIPLET t=quartet[n];

      // lowest weight must be nonnegative

      if (t.A<0.0) {
	cerr << "Error: Quartet file contains negative weight!" << endl;
	exit (1);
      }
      else if (t.A==0.0) {
	t.A=-15.0;
      }
      else {
	t.A=log (t.A);
      }
      
      if (t.B<0.0) {
	cerr << "Error: Quartet file contains negative weight!" << endl;
	exit (1);
      }
      else if (t.B==0.0) {
	t.B=-15.0;
      }
      else {
	t.B=log (t.B);
      }

      if (t.C<0.0) {
	cerr << "Error: Quartet file contains negative weight!" << endl;
	exit (1);
      }
      else if (t.C==0.0) {
	t.C=-15.0;
      }
      else {
	t.C=log (t.C);
      }
      double m=min (min (t.A, t.B), t.C);
      
      t.A-=m;
      t.B-=m;
      t.C-=m;

      quartet[n]=t;
    }
}

double getQuartetWeights (QUARTET quartet, int a, int b, int c, int d)
{
  /**
     Size-order and decide which topology this is
     Access the lists, return the value there
  */
  int x = a;
  int y = b;
  int u = c;
  int v = d;
  int m;
  int idx;
  double result;
  DOUBLETRIPLET w;

  /**
     Create size-ordered quadruple x, y, u, v from a, b, c, d
  */
  if (y>x) 
    {
      m=x;
      x=y;
      y=m;
    }
  if (u>x) 
    {
      m=x;
      x=u;
      u=m;
    }
  if (v>x) 
    {
      m=x;
      x=v;
      v=m;
    }
  if (u>y) 
    {
      m=y;
      y=u;
      u=m;
    }
  if (v>y) 
    {
      m=y;
      y=v;
      v=m;
    }
  if (v>u) 
    {
      m=u;
      u=v;
      v=m;
    }

  /**
     Next, use size-ordered numbers to access the triple
  */
  idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);

  w=quartet[idx];

  /**
     Now, determine which quartet to take. Use the unordered numbers,
     they match one ordering or other
  */
  
  /**
     Result variable
  */
  result = 0.0;
  
  /**
   * If either first or last pair is first or last pair of the stored
   * one, use first weight
  */
  if (   ((a == x || a == y) && (b == x || b == y))
      || ((c == x || c == y) && (d == x || d == y)))
    result = w.A;
  
  /**
   * Next possibility
   */
  else if (   ((a == x || a == u) && (b == x || b == u))
	   || ((c == x || c == u) && (d == x || d == u)))
    result = w.B;
  
  /**
   * Third possibility
   */
  else if (   ((a == x || a == v) && (b == x || b == v))
	   || ((c == x || c == v) && (d == x || d == v)))
    result = w.C;
  
  return (result);
}

void setQuartetWeights (QUARTET& quartet /* reference! */,
			int a, int b, int c, int d,
			double w1, double w2, double w3)
{
  int x = a;
  int y = b;
  int u = c;
  int v = d;
  int m;
  int idx;
  DOUBLETRIPLET w;
  double r1, r2, r3;

  if (y>x)
    {
      m=x;
      x=y;
      y=m;
    }
  if (u>x)
    {
      m=x;
      x=u;
      u=m;
    }
  if (v>x)
    {
      m=x;
      x=v;
      v=m;
    }
  if (u>y)
    {
      m=y;
      y=u;
      u=m;
    }
  if (v>y)
    {
      m=y;
      y=v;
      v=m;
    }
  if (v>u)
    {
      m=u;
      u=v;
      v=m;
    }

  /**
     Investigate which topology of a, b, c, d that the topologies of
     x, y, u, v correspond to, and set weights accordingly
  */

  /**
     See if xy|uv is ab|cd (w1), ac|bd (w2), or ad|bc (w3)
  */

  /**
     The first weight of the stored triplet
  */
  r1=0.0;
  if (   ((x == a || x == b) && (y == a || y == b))
      || ((u == a || u == b) && (v == a || v == b)))
    r1=w1;
  else if (   ((x == a || x == c) && (y == a || y == c))
	   || ((u == a || u == c) && (v == a || v == c)))
    r1=w2;
  else if (   ((x == a || x == d) && (y == a || y == d))
	   || ((u == a || u == d) && (v == a || v == d)))
    r1=w3;
  
  /**
   * See if xu|yv is ab|cd (w1), ac|bd (w2), or ad|bc (w3)
   */

  /**
   * The second weight of the stored triplet
   */
  r2=0.0;
  if (   ((x == a || x == b) && (u == a || u == b))
      || ((y == a || y == b) && (v == a || v == b)))
    r2=w1;
  else if (   ((x == a || x == c) && (u == a || u == c))
	   || ((y == a || y == c) && (v == a || v == c)))
    r2=w2;
  else if (   ((x == a || x == d) && (u == a || u == d))
	   || ((y == a || y == d) && (v == a || v == d)))
    r2=w3;

  /**
   * See if xv|uy is ab|cd (w1), ac|bd (w2), or ad|bc (w3)
   */

  /**
   * The third weight of the stored triplet
   */
  r3=0.0;
  if (   ((x == a || x == b) && (v == a || v == b))
      || ((u == a || u == b) && (y == a || y == b))) 
    r3=w1;
  else if (   ((x == a || x == c) && (v == a || v == c))
	   || ((u == a || u == c) && (y == a || y == c))) 
    r3=w2;
  else if (   ((x == a || x == d) && (v == a || v == d))
	   || ((u == a || u == d) && (y == a || y == d)))
    r3=w3;
  
  w.A=r1;
  w.B=r2;
  w.C=r3;

  /**
     Set to w
  */
  idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);
  quartet[idx]=w;
}

INTVEC
join (INTVEC theFirst, int firstDirection, INTVEC theSecond, int secondDirection)
{
  INTVEC result;

  /* Forward direction of first */
  if (firstDirection == 0) {
    for (INTVEC::const_iterator iter=theFirst.begin(); iter!=theFirst.end(); iter++) {
      result.push_back(*iter);
    }
  }
  /* Backward direction of firs */
  else if (firstDirection == 1) {
    for (INTVEC::const_reverse_iterator iter=theFirst.rbegin(); iter!=theFirst.rend(); iter++) {
      result.push_back(*iter);
    }
  }
  else {
    cout << "Unknown direction of first >" << firstDirection << "<" << endl;
    exit (EXIT_FAILURE);
  }

  /* Forward direction of second */
  if (secondDirection == 0) {
    for (INTVEC::const_iterator iter=theSecond.begin(); iter!=theSecond.end(); iter++) {
      result.push_back(*iter);
    }
  }
  /* Backward direction of second */
  else if (secondDirection == 1) {
    for (INTVEC::const_reverse_iterator iter=theSecond.rbegin(); iter!=theSecond.rend(); iter++) {
      result.push_back(*iter);
    }
  }
  else {
    cout << "Unknown direction of second >" << firstDirection << "<" << endl;
    exit (EXIT_FAILURE);
  }

  return (result);
}

INTINTVEC
join (INTINTVEC theLists, int taxon1, int reversed1, int taxon2, int reversed2)
{
  INTVEC tL1, tL2;

  for (unsigned int n=0; n<theLists.size(); n++) {
    INTVEC tL=theLists[n];
    
    if (count (tL.begin (), tL.end (), taxon1)) tL1=tL;
    if (count (tL.begin (), tL.end (), taxon2)) tL2=tL;
  }
 
  theLists.erase (remove (theLists.begin (), theLists.end (), tL1), theLists.end ());
  theLists.erase (remove (theLists.begin (), theLists.end (), tL2), theLists.end ());

  INTVEC tL12=join (tL1, reversed1, tL2, reversed2);

  theLists.push_back (tL12);

  return theLists;
}

INTINTVEC
join (INTINTVEC theLists, int taxon1, int reversed1, int taxon2, int reversed2, int taxon3, int reversed3)
{
  INTVEC tL1, tL2, tL3;

  for (unsigned int n=0; n<theLists.size(); n++) {
    INTVEC tL=theLists[n];

    if (count (tL.begin (), tL.end (), taxon1)) tL1=tL;
    if (count (tL.begin (), tL.end (), taxon2)) tL2=tL;
    if (count (tL.begin (), tL.end (), taxon3)) tL3=tL;
  }

  theLists.erase (remove (theLists.begin (), theLists.end (), tL1), theLists.end ());
  theLists.erase (remove (theLists.begin (), theLists.end (), tL2), theLists.end ());
  theLists.erase (remove (theLists.begin (), theLists.end (), tL3), theLists.end ());

  INTVEC tL123=join (join (tL1, reversed1, tL2, reversed2), 0, tL3, reversed3);

  theLists.push_back (tL123);

  return theLists;
}

/**
 * Calculate weights of all quadruples
 */
static QUARTET
calcQuartetWeights (vector<string> sequences, string basename, int weights, int silent)
{
  int n;
  int m;
  string input;
  string::const_iterator iter;
  int qsize = over4(sequences.size());

  //std::cerr << "qsize = " << qsize << std::endl;
  /* check if we want more then MAX_INT memory to allocate the array of quartets
     if qsize is smaller then zero we had a memory overflow */
  if (qsize < 0) {
    std::cerr << "Error: number of sequences per alignment exceeded!!!"
	      << std::endl
	      << "       method qnet is O(n^4) in time and O(n^4) in memory."
	      << std::endl
              << "       in terms of taxa this limit is reached at 338 taxa per alignment,"
	      << std::endl
	      << "       which requires about 30GB of main memory !!!!"
	      << std::endl
	      << "       (your input alignment has " << sequences.size() << " taxa)"
	      << std::endl;
    exit (EXIT_FAILURE);
  }

  QUARTET quartets (qsize);
  double s1, s2, s3;
  double w1, w2, w3;
  int a, b, c, d;
  int p;
  unsigned int l, s;
  int patternCount[15];
  char x, y, u, v;
  int cnt = 0; // for progress bar
  ofstream file;

  /* inizialize quartet array */
  for(QUARTET::iterator triplet = quartets.begin(); triplet!=quartets.end(); triplet++) {
    triplet->A = 0.0;
    triplet->B = 0.0;    
    triplet->C = 0.0;
  }
  
  /**
     Check lengths
  */
  n = sequences.size();
  l = sequences[0].length ();
  
  for (m=1; m<n; m++) 
    {
      if ((sequences[m].length ())!=l)
	{
	  cout << "GWeight: Sequences are of different length!" << endl;
	  exit (1);
	}
    }

  if (weights)
    file.open((basename+"_calc.txt").c_str());


  /**
     Then, time for quartets and calculations...	
  */
  if ( ! silent ) cerr << "Calculating quartets       "; 
  for (a=1; a<n-2; a++) {
    // progress bar
    if ( ! silent ) {
      char chars[5] = "|/-\\";
      pretty_print(stderr, "\b\b\b\b\b\b");
      pretty_print(stderr, "%.2f %c",
		   static_cast<float>(cnt)/static_cast<float>(n-3),
                   chars[cnt % 4]);
      cnt++;
      fflush(stderr);
    }
    for (b=a+1;b<n-1; b++) {
      for (c=b+1; c<n; c++) {
	for (d=c+1; d<n+1; d++) {
	  /**
	     Quartet (a, b, c, d)
	  */     
	  
	  /**
	     Null the counts
	  */
	  for (p=0; p<15; p++) patternCount[p]=0;
	  
	  for (s=0; s<l; s++)
	    {
	      /**
		 ... then, check whether the things equal or not...
	      */
	      x=sequences[a-1][s];
	      y=sequences[b-1][s];
	      u=sequences[c-1][s];
	      v=sequences[d-1][s];
	      
	      if (x==y && x==u && x==v)
		patternCount[0]++;
	      else if (y==u && y==v && y!=x)
		patternCount[1]++;
	      else if (y==u && y==v && y!=x)
		patternCount[1]++;
	      else if (x==u && x==v && y!=x) 
		patternCount[2]++;
	      else if (x==y && x==v && x!=u)
		patternCount[3]++;
	      else if (x==y && x==u && x!=v)
		patternCount[4]++;
	      else if (x==y && u==v && x!=u) 
		patternCount[5]++;
	      else if (x==u && y==v && x!=y)
		patternCount[6]++;
	      else if (x==v && y==u && x!=y)
		patternCount[7]++;
	      else if (x==y && u!=v && x!=u && x!=v)
		patternCount[8]++;
	      else if (x==u && x!=y && x!=v && y!=v)
		patternCount[9]++;
	      else if (x==v && x!=y && x!=u && y!=u)
		patternCount[10]++;
	      else if (y==u && y!=x && y!=v && x!=v)
		patternCount[11]++;
	      else if (y==v && y!=x && y!=u && x!=u)
		patternCount[12]++;
	      else if (u==v && x!=u && y!=u && x!=y)
		patternCount[13]++;
	      else if (x!=y && x!=u && x!=v && y!=u && y!=v && u!=v)
		patternCount[14]++;
	    }
	  
	  /**
	     Now we have the counts, create the weights and the triplet
	  */
	  s1=((double)patternCount[5])+0.5*((double)patternCount[8]+(double)patternCount[13]);
	  s2=((double)patternCount[6])+0.5*((double)patternCount[9]+(double)patternCount[12]);
	  s3=((double)patternCount[7])+0.5*((double)patternCount[10]+(double)patternCount[11]);
	  
	  if (s1+s2+s3<0)
	    {
	      cout << "GWeights: Negative weight sum!" << endl;
	      exit (1);
	    }
	  else if (s1+s2+s3==0)
	    {
	      w1=1.0/3.0;
	      w2=1.0/3.0;
	      w3=1.0/3.0;
	    }
	  else 
	    {
	      w1=s1/(s1+s2+s3);
	      w2=s2/(s1+s2+s3);
	      w3=s3/(s1+s2+s3);
	    }
	  
	  /* save weights */
	  setQuartetWeights (quartets, a, b, c, d, w1, w2, w3);

	  // output weights
	  if (weights)
	    print_quartet_to_fstream(file, a, b, c, d, w1, w2, w3);
	}
      }
    }
  }

  /* normalize */
#if LINNORMALIZE
  normalize (quartets);
#endif
#if LOGNORMALIZE
  lognormalize (quartets);
#endif

  if (weights)
    file.close();
  
  if ( ! silent )
    cerr << "\b\b\b\b\b\b" << " done." << endl;

  if (weights)
    print_QuartetWeights_to_file(quartets, basename, sequences.size());
  
  return (quartets);
}

/**
 * Calculate cyclic ordering
 */
INTVEC
qnet_ordering(INTVEC taxa, vector<string> sequences, string basename, int weights, int silent)
{
  double c=0.5;

#if 0
  int cnt = 0; // for progress bar
#endif
  
  /* get number of taxa */
  unsigned int N=taxa.size ();

  /* make N lists each containing one taxa only */
  INTINTVEC theLists;
  for (unsigned int n=0; n<N; n++) theLists.push_back (INTVEC (1, n+1)); // add one here

  /* Calculate quartet weights */
  QUARTET theQuartetWeights=calcQuartetWeights (sequences, basename, weights, silent);
  
  /**
     Method for doing the actual analysis.
     Might return something other than a String, though...
  */
  if ( ! silent )
    cout << "Beginning analysis..." << endl; 

  // begin algorithm as described by Stefan
  // initalization step
  int p=N;

  INTVEC X;
  for (unsigned int n=0; n<N; n++) {
    X.push_back (theLists[n][0]);
  }

  ZHolder zH (theLists, N);
  WHolder wH (theLists, N, theQuartetWeights);
  U0Holder u0H (theLists, N, theQuartetWeights);
  U1Holder u1H (theLists, N, theQuartetWeights);
  NSHolder snH (theLists, N, theQuartetWeights);
  THolder tH (theLists, N, theQuartetWeights);

#if DEBUG
  cout << "After initiation step:" << endl;
  if (X.size ()>2) {
      for (unsigned int xI=0; xI<X.size()-1; xI++) {
	for (unsigned int xJ=xI+1; xJ<X.size(); xJ++) {
	  int i=X[xI];
	  int j=X[xJ];
	  if (snH.getS (i,j)!=0.0) {
	    cout << "n (" << i << ", " << j << "): " << snH.getN (i, j) << endl;
	    cout << "s (" << i << ", " << j << "): " << snH.getS (i, j) << endl;
	  }
	}
	for (unsigned int xJ=0; xJ<X.size(); xJ++) {
	  for (unsigned int xK=0; xK<X.size(); xK++) {
	    int i=X[xI];
	    int j=X[xJ];
	    int k=X[xK];

	    if (tH.getT (i, j, k)!=0.0 && tH.getT (i, j, k)!=0.0) {
	      cout << "t (" << i << ", " << j << ", " << k << "): " << tH.getT (i, j, k) << endl;
	    }
	  }
	}
      }
  }
#endif  

  // iteration step
  while (p>3) {
    // find a, b, a < b, in X so s (a, b) / n (a, b) maximal
    int aMax=-1;
    int bMax=-1;
    double qMax=-100000.00;
    
    for (unsigned int xA=0; xA<X.size()-1; xA++) {
      for (unsigned int xB=xA+1; xB<X.size(); xB++) {
	int a=X[xA];
	int b=X[xB];
	
	double q=snH.getS(a,b)/static_cast<double>(snH.getN(a,b));
	
	if (q>qMax) {
	  qMax=q;
	  aMax=a;
	  bMax=b;
	}
      }
    }
    
    int a=aMax;
    int b=bMax;
    
    // we have found which a, b in X' to join
    // this is equivalent to which lists to join
    
    // the four terms
    
    double yMax=-100000.00;
    int y=-1;
    
    double tABK=0;
    double tBAK=0;
    double tAKB=0;
    double tBKA=0;
    
    for (unsigned int n=0; n<X.size(); n++) {
      int k=X[n];
      
      if (k!=a && k!=b) {
	tABK+=tH.getT(a,b,k);
	tBAK+=tH.getT(b,a,k);
	tAKB+=tH.getT(a,k,b);
	tBKA+=tH.getT(b,k,a);
      }
    }
    
    double y1=(zH.getZ(a)-1)*tABK+(zH.getZ(b)-1)*tBAK+(zH.getZ(b)-1)*(zH.getZ(a)-1)*c*(N-zH.getZ(a)-zH.getZ(b))*u0H.getU(a,b);
    double y2=(zH.getZ(a)-1)*tABK+(zH.getZ(b)-1)*tBKA+(zH.getZ(b)-1)*(zH.getZ(a)-1)*c*(N-zH.getZ(a)-zH.getZ(b))*u1H.getU(a,b);
    double y3=(zH.getZ(a)-1)*tAKB+(zH.getZ(b)-1)*tBAK+(zH.getZ(b)-1)*(zH.getZ(a)-1)*c*(N-zH.getZ(a)-zH.getZ(b))*u1H.getU(a,b);
    double y4=(zH.getZ(a)-1)*tAKB+(zH.getZ(b)-1)*tBKA+(zH.getZ(b)-1)*(zH.getZ(a)-1)*c*(N-zH.getZ(a)-zH.getZ(b))*u0H.getU(a,b);

 #if DEBUG  
    cout << "Deciding on direction to join by:" << endl;
    cout << "sum t (a, b, k): " << tABK << endl;
    cout << "sum t (a, k, b): " << tAKB << endl;
    cout << "sum t (b, a, k): " << tBAK << endl;
    cout << "sum t (b, k, a): " << tBKA << endl;
    cout << "u (a, b, 0): " << u0H.getU (a, b) << endl;
    cout << "u (a, b, 1): " << u1H.getU (a, b) << endl; 
    cout << "y1: " << y1 << " y2: " << y2 << " y3: " << y3 << " y4: " << y4 << endl;
#endif
    
    if (y1>yMax) {
      y=1;
      yMax=y1;
    }
    
    if (y2>yMax) {
      y=2;
      yMax=y2;
    }
    
    if (y3>yMax) {
      y=3;
      yMax=y3;
    }
    
    if (y4>yMax) {
      y=4;
      yMax=y4;
    }

    // we now know which joining to perform
    if (y==1) theLists=join (theLists, a, 1, b, 0);
    else if (y==2) theLists=join (theLists, a, 1, b, 1);
    else if (y==3) theLists=join (theLists, a, 0, b, 0);
    else if (y==4) theLists=join (theLists, a, 0, b, 1);
    else {
      cout << "QNet: y out of range" << endl;
      exit (EXIT_FAILURE);
    }
 
    // remove b from X
    X.erase (remove (X.begin (), X.end (), b), X.end ());
   
    // loop over all elements k in X that are not a, b
    for (unsigned int xK=0; xK<X.size(); xK++) {
      int k=X[xK];
      if (k!=a && k!=b) {
	// modify s
	double sAK=snH.getS(a,k);
	double sBK=snH.getS(b,k);
	double wSum=0.0;
	
	for (unsigned int xL=0; xL<X.size(); xL++) {
	  int l=X[xL];
	  if (l!=a && l!=b && l!=k) {
	    wSum+=wH.getW(a,k,b,l)+wH.getW(b,k,a,l);
	  }
	}
	
	snH.setS(a,k,sAK+sBK-wSum);
	
	// modify n
	int nAK=snH.getN(a,k);
	int nBK=snH.getN(b,k);
	snH.setN (a,k,nAK+nBK-2*(N-zH.getZ(a)-zH.getZ(b)-zH.getZ(k))*zH.getZ(a)*zH.getZ(b)*zH.getZ(k));
	
	for (unsigned int xL=0; xL<X.size(); xL++) {
	  int l=X[xL];

	  // site of troublesome mod

	  if (l!=a && l!=b && l!=k && l<k) {
	    double sKL=snH.getS(k,l);
	    int nKL=snH.getN(k,l);
	    
	    snH.setS(k,l,sKL-wH.getW(a,b,k,l));
	    snH.setN(k,l,nKL-zH.getZ(a)*zH.getZ(b)*zH.getZ(k)*zH.getZ(l));

	    if (y==1) {
	      double m1=tH.getT(a,l,k)+tH.getT(b,k,l)+wH.getW(a,k,b,l);
	      double m2=tH.getT(a,k,l)+tH.getT(b,l,k)+wH.getW(a,l,b,k);
	      tH.setT(a,l,k,m2);
	      tH.setT(a,k,l,m1);
	    }
	    else if (y==2) {
	      double m1=tH.getT(a,l,k)+tH.getT(b,l,k)+wH.getW(a,k,b,l);
	      double m2=tH.getT(a,k,l)+tH.getT(b,k,l)+wH.getW(a,l,b,k);
	      tH.setT(a,l,k,m2);
	      tH.setT(a,k,l,m1);
	    }
	    else if (y==3) {
	      double m1=tH.getT(a,k,l)+tH.getT(b,k,l)+wH.getW(a,k,b,l);
	      double m2=tH.getT(a,l,k)+tH.getT(b,l,k)+wH.getW(a,l,b,k);
	      tH.setT (a,l,k,m2);
	      tH.setT (a,k,l,m1);
	    }
	    else if (y==4) {
	      double m1=tH.getT(a,k,l)+tH.getT(b,l,k)+wH.getW(a,k,b,l);
	      double m2=tH.getT(a,l,k)+tH.getT(b,k,l)+wH.getW(a,l,b,k);
	      tH.setT(a,l,k,m2);
	      tH.setT(a,k,l,m1);
	    }
	    else {
	      cout << "QNet: No y - weird; please report this!" << endl;
	    }

	    double m1=tH.getT(k,a,l)+tH.getT(k,b,l);
	    double m2=tH.getT(l,a,k)+tH.getT(l,b,k);
	    double m3=tH.getT(k,l,a)+tH.getT(k,l,b);
	    double m4=tH.getT(l,k,a)+tH.getT(l,k,b);

	    tH.setT(k,a,l,m1);
	    tH.setT(l,a,k,m2);
	    tH.setT(k,l,a,m3);
	    tH.setT(l,k,a,m4);

	    for (unsigned int xM=0; xM<X.size(); xM++) {
	      int m=X[xM];

	      if (m!=a && m!=b && m!=k && m!=l && m<l) {
		m1=wH.getW(a,k,l,m)+wH.getW(b,k,l,m);
		m2=wH.getW(a,l,k,m)+wH.getW(b,l,k,m);
		m3=wH.getW(a,m,l,k)+wH.getW(b,m,l,k);
		
		wH.setW (a, k, l, m, m1);
		wH.setW (a, l, k, m, m2);
		wH.setW (a, m, l, k, m3);
	      }
	    }
	  }
	}
	    
	if (y==1) {
	  double m1=u1H.getU(a,k)+u0H.getU(b,k)+tH.getT(k,a,b);
	  double m2=u0H.getU(a,k)+u1H.getU(b,k)+tH.getT(k,b,a);
	  u0H.setU(a,k,m1);
	  u1H.setU(a,k,m2);
	}
	
	if (y==2) {
	  double m1=u1H.getU(a,k)+u1H.getU(b,k)+tH.getT(k,a,b);
	  double m2=u0H.getU(a,k)+u0H.getU(b,k)+tH.getT(k,b,a);
	  u0H.setU(a,k,m1);
	  u1H.setU(a,k,m2);
	}
	
	if (y==3) {
	  double m1=u0H.getU(a,k)+u0H.getU(b,k)+tH.getT(k,a,b);
	  double m2=u1H.getU(a,k)+u1H.getU(b,k)+tH.getT(k,b,a);
	  u0H.setU(a,k,m1);
	  u1H.setU(a,k,m2);
	}
	
	if (y==4) {
	  double m1=u0H.getU(a,k)+u1H.getU(b,k)+tH.getT(k,a,b);
	  double m2=u1H.getU(a,k)+u0H.getU(b,k)+tH.getT(k,b,a);
	  u0H.setU(a,k,m1);
	  u1H.setU(a,k,m2);
	}
      }
    }
    
    zH.setZ(a,zH.getZ(a)+zH.getZ(b));
    zH.setZ(b,0);
    
    p--;
    // end of iteration step

#if DEBUG
    if (X.size ()>2) {
      cout << "An iteration just took place. " << X.size () << " paths remaining. Non-zero parameters:" << endl;
      for (unsigned int xI=0; xI<X.size()-1; xI++) {
	for (unsigned int xJ=xI+1; xJ<X.size(); xJ++) {
	  int i=X[xI];
	  int j=X[xJ];
	  if (snH.getS (i, j)!=0.0) {
	    cout << "n (" << i << ", " << j << "): " << snH.getN (i, j) << endl;
	    cout << "s (" << i << ", " << j << "): " << snH.getS (i, j) << endl;
	  }
	  if (u0H.getU (i, j)!=0.0) {
	    cout << "u (" << i << ", " << j << ", 0): " << u0H.getU (i, j) << endl;
	  }
	  if (u1H.getU (i, j)!=0.0) {
	    cout << "u (" << i << ", " << j << ", 1): " << u1H.getU (i, j) << endl;
	  }
	}
	for (unsigned int xJ=0; xJ<X.size(); xJ++) {
	  for (unsigned int xK=0; xK<X.size(); xK++) {
	    int i=X[xI];
	    int j=X[xJ];
	    int k=X[xK];
	    if (tH.getT (i, j, k)!=0.0) {
	      cout << "t (" << i << ", " << j << ", " << k << "): " << tH.getT (i, j, k) << endl;
	    }
	  }
	}
      }
    }
#endif

  }
  
  // termination step
  int i=X[0];
  int j=X[1];
  int k=X[2];
  
  int y=-1;
  double yMax=-100000.00;
  
  double y1=tH.getT(i,k,j)+tH.getT(j,i,k)+tH.getT(k,j,i)+c*(u1H.getU(i,j)+u1H.getU(i,k)+u1H.getU(j,k));
  double y2=tH.getT(i,k,j)+tH.getT(j,i,k)+tH.getT(k,i,j)+c*(u1H.getU(i,j)+u0H.getU(i,k)+u0H.getU(j,k));
  double y3=tH.getT(i,k,j)+tH.getT(j,k,i)+tH.getT(k,j,i)+c*(u0H.getU(i,j)+u1H.getU(i,k)+u0H.getU(j,k));
  double y4=tH.getT(i,k,j)+tH.getT(j,k,i)+tH.getT(k,i,j)+c*(u0H.getU(i,j)+u0H.getU(i,k)+u1H.getU(j,k));
  double y5=tH.getT(i,j,k)+tH.getT(j,i,k)+tH.getT(k,j,i)+c*(u0H.getU(i,j)+u0H.getU(i,k)+u1H.getU(j,k));
  double y6=tH.getT(i,j,k)+tH.getT(j,i,k)+tH.getT(k,i,j)+c*(u0H.getU(i,j)+u1H.getU(i,k)+u0H.getU(j,k));
  double y7=tH.getT(i,j,k)+tH.getT(j,k,i)+tH.getT(k,j,i)+c*(u1H.getU(i,j)+u0H.getU(i,k)+u0H.getU(j,k));
  double y8=tH.getT(i,j,k)+tH.getT(j,k,i)+tH.getT(k,i,j)+c*(u1H.getU(i,j)+u1H.getU(i,k)+u1H.getU(j,k));
  
  if (y1>yMax) {
    y=1;
    yMax=y1;
  }
  
  if (y2>yMax) {
    y=2;
    yMax=y2;
  }
  
  if (y3>yMax) {
    y=3;
    yMax=y3;
  }
  
  if (y4>yMax) {
    y=4;
    yMax=y4;
  }
  
  if (y5>yMax) {
    y=5;
    yMax=y5;
  }
  
  if (y6>yMax) {
    y=6;
    yMax=y6;
  }
  
  if (y7>yMax) {
    y=7;
    yMax=y7;
  }
  
  if (y8>yMax) {
    y=8;
    yMax=y8;
  }

  // we now know which joining to perform
  if (y==1) theLists=join (theLists, i, 0, j, 0, k, 0);
  if (y==2) theLists=join (theLists, i, 0, j, 0, k, 1);
  if (y==3) theLists=join (theLists, i, 0, j, 1, k, 0);
  if (y==4) theLists=join (theLists, i, 0, j, 1, k, 1);
  if (y==5) theLists=join (theLists, i, 1, j, 0, k, 0);
  if (y==6) theLists=join (theLists, i, 1, j, 0, k, 1);
  if (y==7) theLists=join (theLists, i, 1, j, 1, k, 0);
  if (y==8) theLists=join (theLists, i, 1, j, 1, k, 1);

  // end algorithm as described by Stefan
  
  /**
     Now, the lists should contain the desired circular ordering
  */
  
  //cout << "Done." << endl;
  if ( ! silent )
    cerr << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
	 << " done.         " << endl;
 
  INTVEC result;
  for (unsigned int n=0; n<N; n++) 
    result.push_back (theLists[0][n]-1); /* map back from 1,...,N to 0,...,N-1 */
  
  return result;
}

#if 0
int main (int argc, char *argv[])
{
  if (argc<2) {
    cout << "Please specify FASTA file name." << endl;
    exit (EXIT_FAILURE);
  }

  // Read alignment from file
  vector<string> aln=parse_multi_fasta_alignment(argv[1]);
  
  // Calculate cyclic order
  INTVEC order;
  INTVEC inord;
  for (unsigned int i=0; i<aln.size(); i++) inord.push_back(i);
  order=qnet_ordering (inord, aln, 1);

  // Output ordering
  cout << "Ordering: " ;
  copy (order.begin(), order.end(), ostream_iterator<int>(cout, " "));
  cout << endl;

  return (EXIT_SUCCESS);
}
#endif
