/*
  Last changed Time-stamp: <2010-11-11 00:06:17 xtof>
  $Id: FileFormats.cpp,v 1.7 2010/11/10 23:21:01 xtof Exp $
*/
#include <ctime>
#include <fstream>
#include <ios>
#include <iostream>
#include <iterator>
#include <string>
#include <vector>

#include "Maps.h"
#include "TypeDefs.h"
#include "Utils.h"

// private function(s)
static bool
skip_whitespace(std::istream & is);
static bool
parse_word(std::istream & is, std::string & word);
static bool
parse_double(std::istream & is, double & number);
static std::string
PS_main(std::string title);
static std::string
PS_Prolog(std::string);

/**
 * Reads an alignment in multi-fasta format. Returns a pair consisting
 * of an array holding the sequence names and an array holding the
 * rows of the alignment.
 */
Alignment_t
parse_multi_fasta_alignment (char * infile)
{
  std::ifstream file;
  std::vector<std::string> names;
  std::vector<std::string> seqs;

  file.open(infile);

  while (file.good()) {
    std::string line;
    getline(file, line, '\n');

    // skip empty lines
    if ( std::isspace(line[0]) || std::iscntrl(line[0]) )
      continue;

    // found sequence name
    if ( line[0] == '>') {
      names.push_back(fasta_extract_name(line));
      continue;
    }
  
    // found 1st sequence fragment
    if (names.size() > seqs.size()) {
      seqs.push_back(upper_case(trim(line)));
      continue;
    }
    
    // concatenate nth sequence fragment
    // convert line to upper case!
    seqs[seqs.size()-1] += upper_case(trim(line));
  }
  
  file.clear();
  file.close();

  return (make_pair(names, seqs));
}

void
print_multi_fasta_alignment_to_file (Alignment_t aln,
				     std::string filename)
{
  std::ofstream file;
  std::vector<std::string> names = aln.first;
  std::vector<std::string> seqs  = aln.second;
  file.open((filename+"_out.fas").c_str());

  for (unsigned int i=0; i<seqs.size(); ++i) {
    file << "> " << names[i] << std::endl
	 << seqs[i] << std::endl;
  }
  
  file.close();
}

void
print_statistics_to_file(Statistic_t stat,
			 std::vector<std::string> reduced,
			 std::vector<double> smoothed,
			 std::vector<int> order,
			 double rp,
			 double cutoff,
			 char *optStr, std::string basename)
{
  time_t t = time(NULL);
  std::ofstream file;
  std::string c_types = stat.second;
  std::vector<double> p_values = stat.first;
  CharIntMap_t freqs
    = CharIntMap_new_from_char_array(const_cast<char*>(c_types.c_str())); 
  std::string keys = CharIntMap_keys(freqs);
  
  file.open((basename+"_sta.gr").c_str());

  // Print infos
  file << "# time: " << std::asctime(std::localtime(&t))
       << "# commandline: " << optStr << std::endl
       << "# file: " << basename + ".fas" << std::endl
       << "# order: ";

  // Print ordering
  std::copy(order.begin(), order.end(),
	    std::ostream_iterator<int>(file, " "));
  file << std::endl;

  // Print legend
  file << "# legend for statistic:" << std::endl
       << "# N   ... missing data  " << std::endl
       << "# C   ... consonstant sites" << std::endl
       << "# S   ... singleton sites" << std::endl
       << "# P,M ... parsimony informative" << std::endl
       << "# len ... length of original alignment" << std::endl
       << "# red ... lenght of reduced alignment" << std::endl
       << "# RP  ... relative frequency of random sites" << std::endl
       << "# statistic:";

  // Print statistic
  for (std::string::iterator key = keys.begin();
       key != keys.end();
       key++) {
    file << " " << *key << " " << freqs[*key];
  }
  // Print lengths
  file << " " << "len" << " " << p_values.size()
       << " " << "red" << " " << reduced.size()
       << " " << "RP"  << " " << rp
       << std::endl; 

  // Print p-values
  int i = 0;
  for (std::vector<double>::iterator it = p_values.begin();
       it != p_values.end();
       it++) {
    file << ++i << " " << *it << std::endl;
  }
  file << "&" << std::endl;

  // Print smoothed p-values
  if (!smoothed.empty()) {
    i = 0;
    for (std::vector<double>::iterator it = smoothed.begin();
	 it != smoothed.end();
	 it++) {
      file << ++i << " " << *it << std::endl;
    }
    file << "&" << std::endl;
  }

  file.close();
}

void
print_idx_to_file(std::vector<int> keep_idx,
		  std::string basename)
{
  std::ofstream file;
  file.open((basename+"_idx.txt").c_str());

  file << "# col_idx [0,1]" << std::endl;
  for (unsigned int i=0; i<keep_idx.size(); i++) {
    int keep = (keep_idx[i] < 0) ? 0 : 1;
    file << i+1 << " " << keep << std::endl;
  }

  file.close();
}

void
print_type_overview_to_file(Statistic_t stat,
			    std::vector<std::string> taxa,
			    double cutoff,
			    std::string basename)
{
  std::ofstream file;
  std::string c_types = stat.second;
  std::vector<double> p_values = stat.first;

  file.open((basename+"_typ.eps").c_str());

  file << PS_Prolog(basename);

  file << "Guidosdict begin" << std::endl
       << "% data startes here" << std::endl;

  // Print taxa names
  //file << "/TAXA [" << std::endl;
  //for (std::vector<std::string>::iterator it = taxa.begin();
  //     it != taxa.end();
  //     it++) {
  //  file << "(" << *it << ")" << std::endl;
  //}
  //file << "] def" << std::endl
  //     << "/TAXAnum{ TAXA length } bind def" << std::endl;

  // Print column types
  file << "/SEQ (\\" << std::endl;
  unsigned int i = 0;
  while (i<c_types.size()) {
    int len = (i+255 < c_types.size()) ? 255 : c_types.size() - i +1;
    std::string s = c_types.substr(i, len);
    file << s << "\\" << std::endl;
    i += 255;
  }
  file << ") def" << std::endl
       << "/SEQlen { SEQ length } bind def" << std::endl;

  // Print p-values
  file << "/PVALUES [" << std::endl;
  i = 0;
  for (std::vector<double>::iterator it = p_values.begin();
       it != p_values.end();
       it++) {
    file << *it; i++;
    if ((i % 10)==0 ) { file << std::endl; }
    else { file << " "; }
  }
  if ((i % 10)!=0 ) { file << std::endl; }
  file << "] def" << std::endl
       << "/PVALlen { PVALUES length } bind def" << std::endl;

  file << "/CUTOFF " << cutoff << " def" << std::endl
       << "/XDIM 595 def" << std::endl
       << "/YDIM 842 def" << std::endl
       << "/CENTER {XDIM 2 div YDIM 2 div} bind def" << std::endl
       << "/ALNxDIM {XDIM .8 mul SEQlen div} bind def" << std::endl
       << "/TICK {SEQlen log 0.9 sub cvi 50 mul} bind def" << std::endl
       << "/ALNyDIM 100 def" << std::endl

       << "/STR 20 string def" << std::endl;

  file << PS_main(basename+"_typ.eps");

  file.close();
} 


void
print_multi_fasta_alignment (Alignment_t p)
{
  std::vector<std::string> names = p.first;
  std::vector<std::string> seqs  =  p.second;

  for (unsigned int i=0; i<seqs.size(); ++i) {
    pretty_print(stdout, "> %s\n", names[i].c_str());
    pretty_print(stdout, "%s\n", seqs[i].c_str());
  }
}

/**
 * Reade distance matrix of size dim from filename 
 */
DblSqrMatrix_t
read_distance_matrix_from_file(char *filename, int dim)
{
  std::vector<std::vector<double> > D;
  std::ifstream is;

  is.open(filename);
  
  allocate_dbl_sqr_array(D, dim);

  for (int i=0; i<dim; i++) {
    for (int j=0; j<i; j++) {

      if (!parse_double(is, D[i][j]))
	/* print error message */
	;
      /* fill off diagonal elements */
      D[j][i] = D[i][j];
    }

    /* fill diagonal element */
    D[i][i] = 0.0;
  }

  is.close();
  
  return (D);
}

/**
 * Skip whitespace characters from the stream pointed to by is until a
 * word character or EOF is encountered. Returns true upon success,
 * false otherwise.
 */
static bool
skip_whitespace(std::istream & is)
{
  char chr;
  
  while (is) {
    
    if (is.bad() || is.eof()) return (false);

    is.get(chr);
    
    if (std::isspace(chr)) continue;
    is.putback(chr);
    break;
  }
  
  return (true);
}

/**
 * Parse a string value into word from the stream pointed to by
 * is. Leading whitespace is skipped. Returns true upon success, false
 * otherwise.
 */
static bool
parse_word(std::istream & is, std::string & word)
{
  char chr;
  word = "";
  
  /* skip leading whitspace */
  if(!skip_whitespace(is))
    return(false);

  while (is) {

    if (is.bad() || is.eof()) return (false);
     
    is.get(chr);
    
    if (std::isspace(chr)) { /* End of word */
      is.putback(chr);
      return (true);
    }
    
    /* Append chr to word */
    word += chr;
  }

  return (false);
}


/**
 * Parse a double value into number from the stream pointed to by is.
 */
static bool
parse_double(std::istream & is, double & number)
{
  std::string word;

  if (! parse_word(is, word))
    return (false);

  number = strtod(word.c_str(), NULL);

  return (true);
}

static std::string
PS_main(std::string title)
{
  std::string ps_main;
 
  ps_main += "(" + title + ") drawTITLE\n";
  ps_main += "drawALIGNMENT\n";
  ps_main += "markGoodCOLUMNS\n";
  ps_main += "markBadCOLUMNS\n";
  ps_main += "drawSCALE\n";
  ps_main += "showpage\n";
  ps_main += "end\n";
  ps_main += "%%EOF\n";

  return (ps_main);
}

static std::string
PS_Prolog(std::string basename)
{
  std::string prolog;
  std::string timestamp(timeStamp());
  std::string thisfile(__FILE__);

  prolog += "%!PS-Adobe-3.0 EPSF-3.0\n";
  prolog += "%%Title: " + basename + ".fas\n";
  prolog += "%%Creator: " + thisfile + "\n";
  prolog += "%%CreationDate: " + timestamp;
  prolog += "%%BoundingBox: 53 305 543 497\n";
  prolog += "%%DocumentFonts: Helvetica\n";
  prolog += "%%Pages: 1\n";
  prolog += "%%EndComments\n";
  prolog += "%%BeginProlog\n";
  prolog += "/Guidosdict 100 dict def\n";
  prolog += "Guidosdict begin\n";
  prolog += "% color aliases\n";
  prolog += " /col0  {0.000 0.000 0.000 setrgbcolor} bind def % black\n";
  prolog += " /col1  {1.000 0.000 0.000 setrgbcolor} bind def % red\n";
  prolog += " /col10 {0.850 0.000 0.000 setrgbcolor} bind def % darkRed\n";
  prolog += " /col2  {0.000 1.000 0.000 setrgbcolor} bind def % green\n";
  prolog += " /col20 {0.000 0.850 0.000 setrgbcolor} bind def % darkGreen\n";
  prolog += " /col3  {0.000 0.000 1.000 setrgbcolor} bind def % blue\n";
  prolog += " /col31 {0.100 0.900 1.000 setrgbcolor} bind def % lightBlue\n";
  prolog += " /col4  {0.820 0.000 0.820 setrgbcolor} bind def % purple\n";
  prolog += " /col5  {1.000 0.000 1.000 setrgbcolor} bind def % magenta\n";
  prolog += " /col6  {1.000 1.000 0.000 setrgbcolor} bind def % yellow\n";
  prolog += " /col60 {1.000 0.850 0.000 setrgbcolor} bind def % darkYellow\n"; 
  prolog += " /col9  {1.000 1.000 1.000 setrgbcolor} bind def % white\n";

  prolog += "% operator aliases\n";
  prolog += " /C   {col10} bind def % constant  columns\n";
  prolog += " /S   {col1 } bind def % singleton columns\n";
  prolog += " /P   {col20} bind def % parsimony informative\n";
  prolog += " /M   {col2 } bind def % very parsimony informative\n";
  prolog += " /N   {col60} bind def % column with missing data\n";
  prolog += " /cmx {matrix currentmatrix} bind def\n";
  prolog += " /ex  {cvx cvn load exec} bind def\n";
  prolog += " /ff  {findfont} bind def\n";
  prolog += " /gi  {getinterval} bind def\n";
  prolog += " /gr  {grestore} bind def\n";
  prolog += " /gs  {gsave} bind def\n";
  prolog += " /l   {lineto} bind def\n";
  prolog += " /m   {moveto} bind def\n";
  prolog += " /np  {newpath} bind def\n";
  prolog += " /r   {rotate} bind def\n";
  prolog += " /rm  {rmoveto} bind def\n";
  prolog += " /rf  {rectfill} bind def\n";
  prolog += " /scf {scalefont} bind def\n";
  prolog += " /sf  {setfont} bind def\n";
  prolog += " /sh  {show} bind def\n";
  prolog += " /smx {setmatrix} bind def\n";
  prolog += " /stw {stringwidth} bind def\n";
  prolog += " /tr  {translate} bind def\n";

  prolog += "% procedures\n";
  prolog += " /Cshow {gs dup stw pop 2 div neg 0 rm sh gr} bind def\n";
  prolog += " /Rshow {\n";
  prolog += "  gs\n";
  prolog += "  /Helvetica ff 6 scf sf cmx smx -90 r\n";
  prolog += "  fontXDIM 2 div dup neg rm sh\n";
  prolog += "  gr\n";
  prolog += " } bind def\n";
  prolog += " /fontXDIM {(X) stringwidth pop} bind def\n";
  prolog += " /drawTITLE {\n";
  prolog += "  CENTER 1.15 mul m /Helvetica ff 14 scf sf Cshow\n";
  prolog += " } bind def\n";
  prolog += " /drawALIGNMENT {\n";
  prolog += "  gs\n";
  prolog += "  CENTER exch XDIM .8 mul 2 div sub exch ALNyDIM 2 div sub tr\n";
  prolog += "  0 1 SEQlen 1 sub\n";
  prolog += "  { dup SEQ exch 1 gi exch ALNxDIM mul 0\n";
  prolog += "    3 -1 roll gs ex ALNxDIM ALNyDIM rf gr\n";
  prolog += "  } for\n";
  prolog += "  gr\n";
  prolog += " } bind def\n";
  prolog += " /markGoodCOLUMNS {\n";
  prolog += "  gs\n";
  prolog += "  CENTER exch XDIM .8 mul 2 div sub exch\n";
  prolog += "  ALNyDIM 2 div 1.4 mul sub tr\n";
  prolog += "  0 1 PVALlen 1 sub\n";
  prolog += "  { dup PVALUES exch get\n";
  prolog += "    CUTOFF lt\n";
  prolog += "    {pop}\n";
  prolog += "    {ALNxDIM mul 0 ALNxDIM ALNyDIM .15 mul gs col0 rf gr}\n";
  prolog += "    ifelse\n";
  prolog += "  } for\n";
  prolog += "  gr\n";
  prolog += " } bind def\n";
  prolog += " /markBadCOLUMNS {\n";
  prolog += "  gs\n";
  prolog += "  CENTER exch XDIM .8 mul 2 div sub exch\n";
  prolog += "  ALNyDIM 2 div 1.58 mul sub tr\n";
  prolog += "  0 1 PVALlen 1 sub\n";
  prolog += "  { dup PVALUES exch get\n";
  prolog += "    CUTOFF lt\n";
  prolog += "    {ALNxDIM mul 0 ALNxDIM ALNyDIM .15 mul gs col0 rf gr}\n";
  prolog += "    {pop}\n";
  prolog += "    ifelse\n";
  prolog += "  } for\n";
  prolog += "  gr\n";
  prolog += " } bind def\n";
  prolog += " /drawSCALE {\n";
  prolog += "  gs\n";
  prolog += "  CENTER exch XDIM .8 mul 2 div sub exch\n";
  prolog += "  ALNyDIM 2 div 1.8 mul sub tr\n";
  prolog += "  0 1 PVALlen 1 sub\n";
  prolog += "  { dup dup dup 1 add TICK mod 0 eq\n";
  prolog += "    {ALNxDIM mul 0 ALNxDIM ALNyDIM .05 mul gs col0 rf gr\n";
  prolog += "     ALNxDIM mul 0 m 1 add STR cvs Rshow}\n";
  prolog += "    {pop pop pop}\n";
  prolog += "    ifelse\n";
  prolog += "  } for\n";
  prolog += "  gr\n";
  prolog += " } bind def\n";

  prolog += "end\n";
  prolog += "%%EndProlog\n";

  return (prolog);
}
   
// End of file
