/*
  Last changed Time-stamp: <2009-02-06 19:05:35 xtof>
  $Id: Utils.cpp,v 1.10 2011/04/19 16:30:43 xtof Exp $
*/
#include <algorithm>
#include <cerrno>
#include <cstdarg>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <iostream>
#include <iterator>
#include <string>
#include <vector>

#include "Utils.h"

unsigned short xsubi[3];

// Diagnostic functions

void
Fatal(char *fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  fprintf(stderr, "FATAL ERROR: ");
  if ( errno != 0 ) fprintf(stderr, "%s: ", strerror(errno));
  vfprintf(stderr, fmt, args);
  fprintf(stderr, "\n");
  fflush(stderr);
  va_end(args);

  exit(EXIT_FAILURE);
}

void
Warn(char *fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  fprintf(stderr, "WARNING: ");
  vfprintf(stderr, fmt, args);
  fprintf(stderr, "\n");
  fflush(stderr);
  va_end(args);
}

void
error_message(const std::string & message)
{ /* prints error message and aborts */
  std::cerr << std::endl
	    << "ERROR in neighbor net v1.5: "
	    << message << std::endl
	    << "\t Please contact me (David Bryant <bryant@mcb.mcbill.ca>)"
	    << " and tell me what happened" << std::endl;

  exit(EXIT_FAILURE);
}

char *
timeStamp(void)
{
  time_t  cal_time;

  cal_time = time(NULL);
  return ( ctime(&cal_time) );
}

// IO helper functions
void
pretty_print(FILE *hdl, char const *fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  vfprintf(hdl, fmt, args);
  fflush(hdl);
  va_end(args);
}

void
print_int_array(std::vector<int> v)
{
  std::copy(v.begin(), v.end(),
	    std::ostream_iterator<int>(std::cout, " "));
  std::cout << std::endl;
}

void
print_dbl_2Darray(std::vector<std::vector<double> > a)
{
  for (std::vector<std::vector<double> >::iterator it = a.begin();
       it != a.end();
       it++)
    print_dbl_array(*it);
}

void
print_dbl_array(std::vector<double> v)
{
  std::copy(v.begin(), v.end(),
	    std::ostream_iterator<double>(std::cout, " "));
  std::cout << std::endl;
}

void
print_chr_array(std::string s)
{
  std::copy(s.begin(), s.end(),
	    std::ostream_iterator<char>(std::cout, " "));
  std::cout << std::endl;
}

// String helper functions

std::string
format(char const *fmt, ...)
{
  va_list args;
  char tmp[256];
  
  va_start(args, fmt);
  vsprintf(tmp, fmt, args);
  va_end(args);
  
  std::string s(tmp);

  return (s);
}

/**
 * True random shuffling of a list in n-1 steps
 */
std::string
random_shuffle(const std::string & source)
{
  std::string str = source;
  for (unsigned int i = 0; i<str.size(); i++) {
    int p = int_urn(i, str.size()-1);
    std::swap(str[i],str[p]);
  }

  return (str);
}

std::vector<std::string>
split_at(const std::string& source, const std::string& seperator)
{
  std::string str = source;
  std::vector<std::string> pieces;
  size_t position;

  position = str.find_first_of(seperator);
  while (position != str.npos) {
    // don't push empty strings
    if (position != 0) pieces.push_back(str.substr(0, position));
    // remove everything including the seperator
    str.erase(0, position + seperator.length());
    // find next seperator
    position = str.find_first_of(seperator);
  }
  if (str.empty() == false) pieces.push_back(str);

  return (pieces);
}

// function object
struct string_to_integer
  : std::unary_function<std::string, int> {
  int operator()(const std::string s) const {
    return (atoi(s.c_str()));
  }
};

std::vector<int>
to_int_array(std::vector<std::string> strs)
{
  std::vector<int> iarray;
  iarray.resize(strs.size());
  std::transform(strs.begin(),strs.end(), iarray.begin(), string_to_integer());

  return (iarray);  
}

#if 0
split_csv_to_ints(const std::string& source, const std::string& sep)
{
  std::string str = source;
  std::vector<int> iarray;
  size_t pos;

  pos = str.find_first_of(sep);
  while (pos != str.npos) {
    // don't push empty strings
    if (pos != 0) iarray.push_back(atoi(str.substr(0, pos).c_str()));
    // remove everything including the seperator
    str.erase(0, pos + sep.length());
    // find next seperator
    pos = str.find_first_of(sep);
  }
  if (str.empty() == false) iarray.push_back(atoi(str.c_str()));

  return (iarray)
}
#endif

std::string
upper_case(const std::string & source)
{
  std::string str = source;
  transform(str.begin(), str.end(), str.begin(), toupper);

  return (str);
}

std::string
lower_case(const std::string & source)
{
  std::string str = source;
  transform(str.begin(), str.end(), str.begin(), tolower);

  return (str);
}

std::string
trim_right(const std::string & source, const std::string & t)
{
  std::string str = source;

  return (str.erase (str.find_last_not_of(t)+1));
}

std::string
StripPath(const std::string & source, const std::string & t)
{
  std::string str = source;
  return (str.erase(0, str.rfind(t)+1));
}

std::string
StripExtension(const std::string & source, const std::string & t)
{
  std::string str = source;
  str.resize(str.rfind(t));
  return (str);
}

std::string
trim_left(const std::string & source, const std::string & t)
{
  std::string str = source;

  return (str.erase(0, source.find_first_not_of(t)));
}

std::string
trim(const std::string & source, const std::string & t)
{
  std::string str = source;

  return (trim_left(trim_right(str, t), t));
}

std::string
fasta_extract_name(const std::string & source)
{
  std::string str = source;

  return (trim(trim_left(str, ">")));
}

// Nummeric helper function(s)
int
over4 (int n)
{
  if (n > 4) return (((n * (n-1) * (n-2))/6 * (n-3)) / 4);
  else if (n == 4) return (1);
  else return (0);
}

int
over3 (int n)
{
  if (n > 3) return (n * (n-1) * (n-2) / 6 );
  else if (n == 3) return (1);
  else return (0);
}

int
over2 (int n)
{
  if (n > 2) return ( n * (n-1) / 2);
  else if (n == 2) return (1);
  else return (0);
}

int
over1 (int n)
{
  if (n > 1) return (n);
  else if (n == 1) return (1);
  else return 0;
}

// Misc helper function(s)
void
init_rand(void)
{
  time_t t;
  (void) time(&t);
  xsubi[0] = xsubi[1] = xsubi[2] = (unsigned short) t;  /* lower 16 bit */
  xsubi[1] += (unsigned short) ((unsigned)t >> 6);
  xsubi[2] += (unsigned short) ((unsigned)t >> 12);
#ifndef HAVE_ERAND48
  srand((unsigned int) t);
#endif
}

/**
 * uniform random number generator; urn() is in [0,1] uses a linear
 * congruential library routine 48 bit arithmetic
 */
double
urn(void)
{
#ifdef HAVE_ERAND48
  extern double erand48(unsigned short[]);
  return erand48(xsubi);
#else
  return ((double) rand())/RAND_MAX;
#endif
}

int
int_urn(int from, int to)
{
  return (((int)(urn()*(to-from+1))) + from);
}

#if 0
class RandomNumber {
public:
  RandomNumber() {
    //srand(time(0));
    init_rand();
  }
    
  int operator() ( int n) {
    // (int)((double)n * rand()/(RAND_MAX+1.0));
    return int_urn(0, n);
  }
};
#endif

// Memory helper function(s)
void
allocate_dbl_sqr_array(std::vector<std::vector<double> > & A, int size)
{
  A.resize(size);
  for (int i=0; i<size; i++) A[i].resize(size);
}

// End of file
