/*
  Last changed Time-stamp: <2010-11-11 00:08:42 xtof>
  $Id: main.cpp,v 1.11 2010/11/10 23:21:01 xtof Exp $
*/
#include <algorithm>
#include <cstring>
#include <fstream>
#include <iostream>
#include <iterator>
#include <map>
#include <string>
#include <utility>
#include <vector>

#include "options.h"
#include "Alignment.h"
#include "CyclicOrderer.h"
#include "FileFormats.h"
#include "Interfaces.h"
#include "Maps.h"
#include "TypeDefs.h"
#include "Utils.h"

RandomNumber rnd;

int 
factorial (int num)
{
  int result = 1;
  for (int i=1; i<=num; ++i) result *= i;
  return (result);
}

void
all_permutations(Alignment_t aln,
		 CharIntMap_t missing,
		 OptionS* OptS,
		 char *optStr)
{
  time_t t = time(NULL);
  std::ofstream file;
  std::ofstream best;
  std::vector<int> order  = std::vector<int>();
  std::vector<std::pair<int,std::vector<int> > > perms
    =  std::vector<std::pair<int,std::vector<int> > >();

  // initialize the order-array
  for (unsigned int i=0; i<(aln.first).size(); i++) order.push_back(i);
  
  // Rows to columns ...
  std::vector<std::string> cols 
    = listOf_rows_to_listOf_columns(aln.second, order);

  std::string basename(StripPath(StripExtension(OptS->InName)));
  file.open((basename+"_all.gr").c_str());
  best.open((basename+"_best.gr").c_str());

  // Print infos
  file << "# time: " << std::asctime(std::localtime(&t))
       << "# commandline: " << optStr << std::endl
       << "# file: " << basename + ".fas" << std::endl
       << "# permutations: " << factorial((aln.first).size()) << std::endl
       << std::flush;
 
  // process all permutations of the order-array
  int cnt = 0;
  while (next_permutation(order.begin(),order.end())) {

    // Rows to columns ...
    std::vector<std::string> cols 
      = listOf_rows_to_listOf_columns(aln.second, order);
    
    Statistic_t stat = shuffle_each_column(cols, OptS->Shuffles, missing,
					   OptS->NoGap, OptS->Silent,
					   OptS->InName);
    
    double rp = relative_frequency_of_random_columns(cols, stat, OptS->CutOff);
    // remember permutations with a "good" RP value
    if (rp <= OptS->CutOff) {
      best << "# ";
      std::copy(order.begin(), order.end(),
		std::ostream_iterator<int>(best, " "));
      best << std::endl;
      best << cnt+1 << " " << rp << std::endl << std::flush;
    }
    file << ++cnt << " " << rp << std::endl << std::flush;
  }

  file.close();
  best.close();
}

void
random_permutations(Alignment_t aln,
		    CharIntMap_t missing,
		    OptionS* OptS,
		    char *optStr)
{
  time_t t = time(NULL);
  std::ofstream file;
  std::ofstream best;
  std::vector<int> order  = std::vector<int>();
  std::vector<std::pair<int,std::vector<int> > > perms
    =  std::vector<std::pair<int,std::vector<int> > >();

  // initialize the order-array
  for (unsigned int i=0; i<(aln.first).size(); i++) order.push_back(i);
  
  // Rows to columns ...
  std::vector<std::string> cols 
    = listOf_rows_to_listOf_columns(aln.second, order);

  std::string basename(StripPath(StripExtension(OptS->InName)));
  file.open((basename+"_rand.gr").c_str());
  best.open((basename+"_best.gr").c_str());
  
  // Print infos
  file << "# time: " << std::asctime(std::localtime(&t))
       << "# commandline: " << optStr << std::endl
       << "# file: " << basename + ".fas" << std::endl
       << "# sample size: " << OptS->SampleSize << std::endl
       << std::flush;
 
  // process SampleSize random permutations of the order-array
  for (int cnt=1; cnt <= OptS->SampleSize; cnt++) {
    std::random_shuffle(order.begin(), order.end(), rnd);

    // Rows to columns ...
    std::vector<std::string> cols 
      = listOf_rows_to_listOf_columns(aln.second, order);
    
    Statistic_t stat = shuffle_each_column(cols, OptS->Shuffles, missing,
					   OptS->NoGap, OptS->Silent,
					   OptS->InName);
    
    double rp = relative_frequency_of_random_columns(cols, stat, OptS->CutOff);
    // print permutations with a "good" RP value to file
    if (rp <= OptS->CutOff) {
      best << "# ";
      std::copy(order.begin(), order.end(),
		std::ostream_iterator<int>(best, " "));
      best << std::endl;
      best << cnt << " " << rp << std::endl << std::flush;
    }
    file << cnt << " " << rp << std::endl << std::flush;
  }

  file.close();
  best.close();
}

int
main (int argc, char *argv[])
{
  // Process commandline options
  OptionS* OptS = decodeCML(argc, argv);

  // generate basename
  std::string basename(StripPath(StripExtension(OptS->InName)));
  
  // Make missing hash
  CharIntMap_t missing = CharIntMap_new_from_char_array(OptS->Missing);

  // Read alignment from file
  Alignment_t aln = parse_multi_fasta_alignment(OptS->InName);
  
  if (OptS->Verbose >=2) print_multi_fasta_alignment(aln);

  // Read distance matrix from file
  std::vector<std::vector<double> > D;
  if (OptS->DistMx[0] != '\0')
    D = read_distance_matrix_from_file(OptS->DistMx, aln.second.size());

  // Calculate cyclic order
  std::vector<int> order;
  if (strcmp(OptS->OrdMethod, "nnet") == 0)
    order = nnet_ordering(aln.second, D,
			  OptS->SeqType, OptS->DistMethod);
  else if (strcmp(OptS->OrdMethod, "qnet") == 0) {
    std::vector<int> inord = std::vector<int>();
    for (unsigned int i=0; i<(aln.first).size(); i++) inord.push_back(i);
#ifdef _QNET_WEIGHTS_DEBUG_    
    order = qnet_ordering(inord, aln.second, basename, 1, OptS->Silent);
#else
    order = qnet_ordering(inord, aln.second, basename, 0, OptS->Silent);
#endif
  }
  else if (strncmp(OptS->OrdMethod, "rand", 4) == 0) {
    if (strchr(OptS->OrdMethod, ',') != NULL) {
      std::vector<std::string> s = split_at(OptS->OrdMethod);
      OptS->SampleSize = atoi(s[1].c_str());
    }
    random_permutations(aln, missing, OptS, optionString());
    exit (EXIT_SUCCESS);
  }
  else if (strcmp(OptS->OrdMethod, "all") == 0) {
    std::cerr << " Calculating RP value for all permutations "
	      << factorial((aln.first).size())
	      << std::endl
	      << " Stay calm, this may take a while!"
	      << std::endl;
    all_permutations(aln, missing, OptS, optionString());
    exit (EXIT_SUCCESS);
  }
  else if (strchr(OptS->OrdMethod, ',') != NULL) {
    order = to_int_array(split_at(OptS->OrdMethod));
  }
  else {
    std::cout << "ERROR: Unknown ordering method `" << OptS->OrdMethod
	      << "'" << std::endl;
    exit (EXIT_FAILURE);
  }

  if (OptS->Verbose) print_dbl_2Darray(D);
  if (OptS->Verbose >= 1) print_int_array(order);

  // Reorder input alignment only
  if (OptS->ReOrder == 1) {
    print_multi_fasta_alignment(make_pair(to_cyclic_order(aln.first, order),
					  to_cyclic_order(aln.second,order)));
    exit (EXIT_SUCCESS);
  }

  // Rows to columns ...
  std::vector<std::string> cols 
    = listOf_rows_to_listOf_columns(aln.second, order);

  // Do the column shuffling and classify column types
  Statistic_t stat = shuffle_each_column(cols, OptS->Shuffles, missing,
					 OptS->NoGap, OptS->Silent,
					 OptS->InName);
  
  // Calculate running average
  std::vector<double> smoothed = smooth_values(stat.first, OptS->Smooth);
  
  // Output in verbose mode
  if (OptS->Verbose >= 1) print_dbl_array(stat.first);
  if (OptS->Verbose >= 1) print_chr_array(stat.second);

  // Remove constant and singleton columns and collums with p-value
  // below cutoff
  Result_t res = filter_columns(cols, stat, OptS->CutOff, OptS->NoConstant);
  std::vector<std::string> reduced = res.first;
  std::vector<int> keep_idx = res.second;

  print_idx_to_file(keep_idx, basename);
  
  // Revert list of alignment columns to list of alignment rows
  std::vector<std::string> rows = listOf_columns_to_listOf_rows(reduced);
  
  print_multi_fasta_alignment_to_file(make_pair(to_cyclic_order(aln.first,
								order),
						rows), basename);

  double rp = relative_frequency_of_random_columns(cols, stat, OptS->CutOff);
  
  print_statistics_to_file(stat, reduced, smoothed, order,
			   rp, OptS->CutOff, optionString(), basename);
  

  print_type_overview_to_file(stat, aln.first, OptS->CutOff, basename);

  return (EXIT_SUCCESS);
}


// End of file
