/*******************************************************
filename: 	matrix_analysis.cpp
author: 	David Bryant (bryant@mcb.mcgill.ca)
created: 	JUNE 2003
copyright:	David Bryant 2003

********************************************************/

#include <cmath>

#include "matrix_analysis.h"

/* resize

Resize equivalents for 1d and 2d arrays. */
void resize(Array2D& A, int n) {
    int i;
    A.resize(n);
    for(i=0;i<n;i++)
        A[i].resize(n);
}


void resize(Array1D& v, int n) {
    v.resize(n);
}


/**********************
TriDiagonalReduction

Conjecture: TriDiagonal form of V stored in vectors d [diagonal] and e [off-digaonal]
The matrix V is replace by an orthogonal matrix Q such that Q^T V Q = T, T tridiagonal.

*/

static void TriDiagonalReduction(Array2D& V, Array1D& d, Array1D& e) {

    int i,j,k,n;
    double scale,h,hh,f,g;
    
    n = V.size();
    
    for (int j = 0; j < n; j++) {
        d[j] = V[n-1][j];
    }

    for (int i = n-1; i > 0; i--) {

        /* Scale to avoid under/overflow.*/

        scale = 0.0;
        h = 0.0;
        for (k = 0; k < i; k++) {
            scale += ABS(d[k]);
        }
        if (scale == 0.0) {
            e[i] = d[i-1];
            for (int j = 0; j < i; j++) {
                d[j] = V[i-1][j];
                V[i][j] = 0.0;
                V[j][i] = 0.0;
            }
        }
        else {

            /* Generate Householder vector for the (scaled) vector d[0..i-1] */

            for (k = 0; k < i; k++) {
                d[k] /= scale;
                h += d[k] * d[k];
            }
            f = d[i-1];
            g = sqrt(h);
            if (f > 0) {
                g = -g;
            }
            e[i] = scale * g;
            h = h - f * g;
            d[i-1] = f - g;
            for (j = 0; j < i; j++) {
                e[j] = 0.0;
            }

            // Apply similarity transformation to remaining columns.

            for (int j = 0; j < i; j++) {
                f = d[j];
                V[j][i] = f;
                g = e[j] + V[j][j] * f;
                for (int k = j+1; k <= i-1; k++) {
                    g += V[k][j] * d[k];
                    e[k] += V[k][j] * f;
                }
                e[j] = g;
            }
            f = 0.0;
            for (int j = 0; j < i; j++) {
                e[j] /= h;
                f += e[j] * d[j];
            }
            hh = f / (h + h);
            for (int j = 0; j < i; j++) {
                e[j] -= hh * d[j];
            }
            for (int j = 0; j < i; j++) {
                f = d[j];
                g = e[j];
                for (int k = j; k <= i-1; k++) {
                    V[k][j] -= (f * e[k] + g * d[k]);
                }
                d[j] = V[i-1][j];
                V[i][j] = 0.0;
            }
        }
        d[i] = h;
    }
   
    // Accumulate transformations.
   
    for ( i = 0; i < n-1; i++) {
        V[n-1][i] = V[i][i];
        V[i][i] = 1.0;
        h = d[i+1];
        if (h != 0.0) {
            for ( k = 0; k <= i; k++) {
                d[k] = V[k][i+1] / h;
            }
            for ( j = 0; j <= i; j++) {
                g = 0.0;
                for ( k = 0; k <= i; k++) {
                    g += V[k][i+1] * V[k][j];
                }
                for ( k = 0; k <= i; k++) {
                    V[k][j] -= g * d[k];
                }
            }
        }
        for ( k = 0; k <= i; k++) {
            V[k][i+1] = 0.0;
        }
    }
    for ( j = 0; j < n; j++) {
        d[j] = V[n-1][j];
        V[n-1][j] = 0.0;
    }
    V[n-1][n-1] = 1.0;
    e[0] = 0.0;
} 

/****************
TriDiagonalQL

Symmetric TriDiagonal QR algorithm
****************/

static void TriDiagonalQR(Array2D& V, Array1D& d, Array1D& e) {

    //  This is derived from the Algol procedures tql2, by
    //  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
    //  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
    //  Fortran subroutine in EISPACK.
    int i,l,n;

    double f,tst1,eps;
    double g,p,r,dl1,h;
    double c,c2,c3,el1,s,s2;

    n = V.size();
    
    for (i = 1; i < n; i++) {
        e[i-1] = e[i];
    }
    e[n-1] = 0.0;
   
    f = 0.0;
    tst1 = 0.0;
    eps = pow(2.0,-52.0);
    for (l = 0; l < n; l++) {

        // Find small subdiagonal element

      tst1 = std::max(tst1,ABS(d[l]) + ABS(e[l]));
        int m = l;

        // Original while-loop from Java code
        while (m < n) {
            if (ABS(e[m]) <= eps*tst1) {
                break;
            }
            m++;
        }


        // If m == l, d[l] is an eigenvalue,
        // otherwise, iterate.

        if (m > l) {
            int iter = 0;
            do {
                iter = iter + 1;  // (Could check iteration count here.)

                // Compute implicit shift

                g = d[l];
                p = (d[l+1] - g) / (2.0 * e[l]);
                r = hypot(p,1.0);
                if (p < 0) {
                    r = -r;
                }
                d[l] = e[l] / (p + r);
                d[l+1] = e[l] * (p + r);
                dl1 = d[l+1];
                h = g - d[l];
                for (int i = l+2; i < n; i++) {
                    d[i] -= h;
                }
                f = f + h;

                // Implicit QL transformation.



                p = d[m];
                c = 1.0;
                c2 = c;
                c3 = c;
                el1 = e[l+1];
                s = 0.0;
                s2 = 0.0;
                for (int i = m-1; i >= l; i--) {
                    c3 = c2;
                    c2 = c;
                    s2 = s;
                    g = c * e[i];
                    h = c * p;
                    r = hypot(p,e[i]);
                    e[i+1] = s * r;
                    s = e[i] / r;
                    c = p / r;
                    p = c * d[i] - s * g;
                    d[i+1] = h + s * (c * g + s * d[i]);

                    // Accumulate transformation.

                    for (int k = 0; k < n; k++) {
                        h = V[k][i+1];
                        V[k][i+1] = s * V[k][i] + c * h;
                        V[k][i] = c * V[k][i] - s * h;
                    }
                }
                p = -s * s2 * c3 * el1 * e[l] / dl1;
                e[l] = s * p;
                d[l] = c * p;

                // Check for convergence.

            } while (ABS(e[l]) > eps*tst1);
        }
        d[l] = d[l] + f;
        e[l] = 0.0;
    }

    // Sort eigenvalues and corresponding vectors.
   
    for (int i = 0; i < n-1; i++) {
        int k = i;
        p = d[i];
        for (int j = i+1; j < n; j++) {
            if (d[j] < p) {
                k = j;
                p = d[j];
            }
        }
        if (k != i) {
            d[k] = d[i];
            d[i] = p;
            for (int j = 0; j < n; j++) {
                p = V[j][i];
                V[j][i] = V[j][k];
                V[j][k] = p;
            }
        }
    }
}   



/**********************
Diagonalise

Returns Diagonalisation of the matrix A, using the QR Symmetric Householder algorithm.
The matrix A is replace by an orthonormal matrix of eigenvectors for A. The eigenvalues are
returned in the vectors d [real part] and e [imaginary part], with ordering corresponding to 
the ordering of e-vectors in V.

Adapted from jama_eig.h in JAMA package

http://math.nist.gov/tnt/index.html
 */

bool Diagonalise(Array2D& A, Array1D& d, Array1D& e) {
    
    int i,j,n;
    bool is_symmetric;

    n = A.size();
    d.resize(n);
    e.resize(n);
    
    is_symmetric = true;
    for(i=0; i<n && is_symmetric ; i++)
        for(j=0;j<n && is_symmetric; j++) 
            is_symmetric = (A[i][j] == A[j][i]);
    
    if (!is_symmetric)
        return false;

    TriDiagonalReduction(A,d,e);
    TriDiagonalQR(A,d,e);

    return true;
    
}
