/*******************************************************
filename: 	neighbor_net.cpp
author: 	David Bryant (bryant@math.mcgill.ca)
created: 	June 2001
copyright:	David Bryant 2001 

********************************************************/

#include <list>

#include"neighbor_net.h"

class net_node {
public:
  net_node() : id(-1), nbr(NULL), ch1(NULL),ch2(NULL), prev(NULL), next(NULL), Rx(0.0), Sx(0.0) {}
  ~net_node() {}
  int id;
  net_node *nbr; /* Adjacent node */
  net_node *ch1,*ch2; /* Child nodes */
  net_node *prev, *next; /* Next and previous active nodes */
  double Rx; /* Sum of distances from this node to all active nodes */
  double Sx; /* Sum of the average distances between this cluster and the other clusters.*/
};


typedef std::vector< std::vector< double > > dbl_array;

/***************
Agglomeration routines

***************/

static void agg2way(net_node* x, net_node* y) {
  //printf("Two way: %d, %d\n",x->id,y->id); 

  x->nbr=y;
  y->nbr=x;
}

static net_node* agg3way(net_node* x,
			 net_node* y,
			 net_node* z, 
			 std::list<net_node*>& amalgs,
			 dbl_array& D, net_node* NetNodes,
			 int& num_nodes) {
  /* Agglomerate x,y, and z to give TWO new nodes, u and v */
  /* In terms of the linked list: we replace x and z 
	 by u and v and remove y from the linked list.
	 and replace y with the new node z 
  Returns a pointer to the node u */
    //printf("Three way: %d, %d, and %d\n",x->id,y->id,z->id); 

  net_node* u = new net_node;
  u->id = num_nodes;
  num_nodes++;
  u->ch1 = x;
  u->ch2 = y;
  
  net_node* v = new net_node;
  v->id = num_nodes;
  num_nodes++;
  v->ch1 = y;
  v->ch2 = z;

  /* Replace x by u in the linked list */
  u->next = x->next;
  u->prev = x->prev;
  if (u->next!=NULL)
	u->next->prev = u;
  if (u->prev!=NULL)
	u->prev->next = u;

  /* Replace z by v in the linked list */
  v->next = z->next;
  v->prev = z->prev;
  if (v->next!=NULL)
	v->next->prev = v;
  if (v->prev!=NULL)
	v->prev->next = v;
  
  /* Remove y from the linked list */
  if (y->next!=NULL)
	y->next->prev = y->prev;
  if (y->prev!=NULL)
	y->prev->next = y->next;

  /* Add an edge between u and v, and add u into the list of amalgamations */
  u->nbr=v;
  v->nbr=u;

  /* Update distance matrix */

  for(net_node* p = NetNodes->next;p!=NULL;p=p->next) {
	D[u->id][p->id] = D[p->id][u->id]=(2.0/3.0)* D[x->id][p->id] + (1.0/3.0)*D[y->id][p->id];
        D[v->id][p->id] = D[p->id][v->id]=(2.0/3.0)* D[z->id][p->id] + (1.0/3.0)*D[y->id][p->id];
  }
  D[u->id][u->id]=D[v->id][v->id]=0.0;
  
  amalgs.push_back(u);

  return u;
}

static void agg4way(net_node* x2,
		    net_node* x,
		    net_node* y,
		    net_node* y2, 
                    std::list<net_node*>& amalgs,
		    dbl_array& D,
		    net_node* NetNodes,int& num_nodes) {
  /* Replace x2,x,y,y2 by with two vertices... performed using two 
	 3 way amalgamations */

  net_node* u;

  u = agg3way(x2,x,y,amalgs,D,NetNodes,num_nodes); /* Replace x2,x,y by two nodes, equal to x2_prev->next and y_prev->next. */
  agg3way(u,u->nbr,y2,amalgs,D,NetNodes,num_nodes); /* z = y_prev -> next */
}


/*********************************************
agglom_nodes

This is the guts of the method: an agglomerative process for constructing a cyclic
network. 
********************************************/

static double compute_Rx(net_node *z, net_node *Cx, net_node *Cy, const dbl_array& D, net_node* NetNodes) {
    double Rx = 0.0;
    net_node *p;
    
    for (p=NetNodes->next; p!=NULL; p = p->next) {
        if (p==Cx || p==Cx->nbr || p == Cy || p == Cy->nbr || p->nbr == NULL) 
            Rx += D[z->id][p->id];
        else /* p->nbr != NULL */
            Rx += D[z->id][p->id]/2.0; /* We take the average of the distances */
    }
    
    return Rx;
    
}

static void agglom_nodes(std::list<net_node*>& amalgs, dbl_array& D, net_node* NetNodes, int& num_nodes) {

    net_node *p,*q, *Cx,*Cy,*x, *y;
    double Qpq,best;
    int num_active = num_nodes;
    int num_clusters = num_nodes;
    int m;
    double Dpq;
    
    while (num_active>3) {
	
        /* Special case
        If we let this one go then we get a divide by zero when computing Qpq */
        if (num_active==4 && num_clusters==2) {
            p = NetNodes->next;
            if (p->next != p->nbr)
                q=p->next;
            else
                q=p->next->next;
            if (D[p->id][q->id] + D[p->nbr->id][q->nbr->id] < D[p->id][q->nbr->id] + D[p->nbr->id][q->id])
                agg3way(p,q,q->nbr,amalgs,D,NetNodes,num_nodes);
            else
                agg3way(p,q->nbr,q,amalgs,D,NetNodes,num_nodes);
            break;
        }
        
        
        /* Compute the "averaged" sums s_i from each cluster to every other cluster. 
        
        To Do: 2x speedup by using symmetry*/
        
        for (p=NetNodes->next; p!=NULL; p = p->next) 
            p->Sx = 0.0;
        for (p=NetNodes->next; p!=NULL; p = p->next) {
            if (p->nbr==NULL || p->nbr->id > p->id ) {
                for (q=p->next;q!=NULL;q = q->next) {
		  if (q->nbr == NULL || ((q->nbr->id > q->id)&&(q->nbr!=p))) { 
                        Dpq = 0.0;
                        if ((p->nbr == NULL) && (q->nbr == NULL)) 
                            Dpq = D[p->id][q->id];
                        else if  ((p->nbr != NULL) && (q->nbr == NULL))
                            Dpq = (D[p->id][q->id] + D[p->nbr->id][q->id])/2.0;
                        else if  ((p->nbr == NULL) && (q->nbr != NULL))
                            Dpq = (D[p->id][q->id] + D[p->id][q->nbr->id])/2.0;
                        else
                            Dpq = (D[p->id][q->id] + D[p->id][q->nbr->id] + D[p->nbr->id][q->id] + D[p->nbr->id][q->nbr->id])/4.0;
                    
                        p->Sx+=Dpq;
                        if (p->nbr!=NULL)
                            p->nbr->Sx += Dpq;
                        q->Sx+=Dpq;
                        if (q->nbr!=NULL)
                            q->nbr->Sx += Dpq;
                    }
                }
            }
        }
    
        Cx = Cy = NULL;
        /* Now minimize (m-2) D[C_i,C_k] - Sx - Sy */
        best=10000000;
        for (p=NetNodes->next; p!=NULL; p = p->next) {
            if ((p->nbr!=NULL) && (p->nbr->id < p->id)) /* We only evaluate one node per cluster */
                continue;
            for (q=NetNodes->next;q!=p;q = q->next) {
                if ((q->nbr!=NULL) && (q->nbr->id < q->id)) /* We only evaluate one node per cluster */
                    continue;
                if (q->nbr == p) /* We only evaluate nodes in different clusters */
                    continue;
                if ((p->nbr == NULL) && (q->nbr == NULL)) 
                    Dpq = D[p->id][q->id];
                else if  ((p->nbr != NULL) && (q->nbr == NULL))
                    Dpq = (D[p->id][q->id] + D[p->nbr->id][q->id])/2.0;
                else if  ((p->nbr == NULL) && (q->nbr != NULL))
                    Dpq = (D[p->id][q->id] + D[p->id][q->nbr->id])/2.0;
                else
                    Dpq = (D[p->id][q->id] + D[p->id][q->nbr->id] + D[p->nbr->id][q->id] + D[p->nbr->id][q->nbr->id])/4.0;
                Qpq = ((double)num_clusters - 2.0) * Dpq - p->Sx - q->Sx;
                /* Check if this is the best so far */
                if ((Cx==NULL || (Qpq < best))&&(p->nbr!=q)) {
                    Cx = p;
                    Cy = q;
                    best = Qpq;
                }
            }
        }
        
        
        
        /* Find the node in each cluster */
        x = Cx;
        y = Cy;

        if (Cx->nbr!=NULL || Cy->nbr != NULL) {
            Cx->Rx = compute_Rx(Cx,Cx,Cy,D,NetNodes);
            if (Cx->nbr != NULL)
                Cx->nbr->Rx = compute_Rx(Cx->nbr,Cx,Cy,D,NetNodes);
            Cy->Rx = compute_Rx(Cy,Cx,Cy,D,NetNodes);
            if (Cy->nbr != NULL)
                Cy->nbr->Rx = compute_Rx(Cy->nbr,Cx,Cy,D,NetNodes);
        }
            
        m = num_clusters;
        if (Cx->nbr != NULL) 
            m++;
        if (Cy->nbr != NULL)
            m++;
            
        best = ((double)m - 2.0) * D[Cx->id][Cy->id] - Cx->Rx - Cy->Rx;
        if (Cx->nbr != NULL) {
            Qpq = ((double)m - 2.0) * D[Cx->nbr->id][Cy->id] - Cx->nbr->Rx - Cy->Rx;
            if (Qpq < best) {
                x = Cx->nbr;
                y = Cy;
                best = Qpq;
            }
        }
        if (Cy->nbr != NULL) {
            Qpq = ((double)m - 2.0) * D[Cx->id][Cy->nbr->id] - Cx->Rx - Cy->nbr->Rx;
            if (Qpq < best) {
                x = Cx;
                y = Cy->nbr;
                best = Qpq;
            }
        }
        if ((Cx->nbr != NULL) && (Cy->nbr != NULL)) {
            Qpq = ((double)m - 2.0) * D[Cx->nbr->id][Cy->nbr->id] - Cx->nbr->Rx - Cy->nbr->Rx;
            if (Qpq < best) {
                x = Cx->nbr;
                y = Cy->nbr;
                best = Qpq;
            }
        }
        
      //  cout<<"Agglomerating "<<x->id<<" and "<<y->id<<endl;
        

        
        /* We perform an agglomeration... one of three types */
        if ((NULL==x->nbr)&&(NULL==y->nbr)) {   /* Both vertices are isolated...add edge {x,y} */
            agg2way(x,y);
            num_clusters--;
        }
        else if (NULL==x->nbr) { 	/* X is isolated,  Y  is not isolated*/
            agg3way(x,y,y->nbr,amalgs,D,NetNodes,num_nodes);
            num_active--;
            num_clusters--;
        }
        else if ((NULL==y->nbr)||(num_active==4)) { /* Y is isolated,  X is not isolated  
                                                    OR theres only four active nodes and none are isolated */
            agg3way(y,x,x->nbr,amalgs,D,NetNodes,num_nodes);
            num_active--;
            num_clusters--;
        }
        else {  /* Both nodes are connected to others and there are more than 4 active nodes */
            agg4way(x->nbr,x,y,y->nbr,amalgs,D,NetNodes,num_nodes);
            num_active-=2;
            num_clusters--;
        }
     }  
}


static void quick_expand(int num_nodes,
                        int ntax,
                        std::list<net_node*>& amalgs, 
                        net_node* NetNodes,
                        std::vector<int>& ordering
                        ) {

    net_node *x,*y,*z,*u,*v, *a;
    
    /* Set up the circular order for the first three nodes */
    x = NetNodes->next; 
    y = x->next;
    z = y->next;
    z->next = x; 
    x->prev = z;
  
    /* Now do the rest of the expansions */
    while (!amalgs.empty()) {
        /* Find the three elements replacing u and v. Swap u and v around if v comes before u in the
        circular ordering being built up */
        u=amalgs.back();
        v=u->nbr;
        x = u->ch1;
        y = u->ch2;
        z = v->ch2;
        if (v!=u->next) { 
            std::swap(u,v);
            std::swap(x,z);
        }

    
        /* Insert x,y,z into the circular order */
        x->prev = u->prev; x->prev->next = x;
        x->next = y; y->prev = x;
        y->next = z; z->prev = y;
        z->next = v->next; z->next->prev = z;
    
        delete u;
        delete v;

	amalgs.pop_back();
    
    }

    /* When we exit, we know that the point x points to a node in the circular order */
    /* We loop through until we find the node after taxa zero */
    while(x->id!=0) {
        x=x->next;
    }

/* extract the ordering */
    ordering.clear();
    a = x;
    do {
        ordering.push_back(a->id);
        a = a->next;
    } while (a!=x);

/* free up the linked list */
    a = x -> next;
    while(a!=x) {
        a = a->next;
        delete a->prev;
    }
    delete x;
    
}

void neighbor_net (const dbl_array& inputD, std::vector<int>& ordering) {

    net_node* NetNodes;
    int num_nodes,ntax;
    int i,j;
    dbl_array D;
  
    /* Initialise data structures */
    ntax = inputD.size();
    
    num_nodes = ntax; /*Initially start with one node for each taxa */
    int max_num_nodes = ntax*3-6;
    /* Copy the distance matrix into a larger, scratch distance matrix */
    D.resize(max_num_nodes); //To Do... check memory
    for(i=0;i<max_num_nodes;i++)
        D[i].resize(max_num_nodes);
    for(j=0;j<ntax;j++) {
        D[j][j] = 0.0;
        for(i=0;i<j;i++) 
            D[i][j]=D[j][i]=inputD[j][i];
    }
    
    

    NetNodes = new net_node;
    net_node* tax_node;

    /* Nodes are stored in a doubly linked list that we set up here */
    for(i=ntax-1;i>=0;i--) { /* Initially, all singleton nodes are active */
	tax_node = new net_node;
	tax_node->id = i;
	tax_node->next=NetNodes->next;
	NetNodes->next=tax_node;
    }
    for (tax_node=NetNodes;tax_node->next!=NULL;tax_node = tax_node->next)   /* Set up links in other direction */
	tax_node->next->prev = tax_node;

	//cerr<<"Set up NNet"<<endl;

    /* Perform the agglomeration step */
    std::list<net_node*> amalgs;
    agglom_nodes(amalgs,D,NetNodes,num_nodes);
    
    	//cerr<<"Done aggloms"<<endl;

    quick_expand(num_nodes,ntax,amalgs,NetNodes,ordering);
    
    
}

